# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
# (C) Chalmers Next Labs (2025)
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# ---------------- ALTERATION NOTICE ---------------- #
# This code has been derived from the Qiskit json encoder in the IBMQ providers github.
# The reason is to remove IBMQ as a dependency, but still be able to use the functionality.
# It has not been altered.
#
# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#

"""Handles the JSON serialization of job objects"""
import json
from typing import Any

from qiskit.circuit.parameterexpression import ParameterExpression


class IQXJsonEncoder(json.JSONEncoder):
    """A json encoder for qobj"""

    def __encode(self, param: Any) -> Any:
        """
        Convert dictionary to contain only JSON serializable types. For example,
        if the key is a Parameter we convert it to a string.
        """
        if isinstance(param, dict):
            param_bind_str = {}
            for key in param.keys():
                value = self.__encode(param[key])

                if isinstance(key, (bool, float, int, str)) or key is None:
                    param_bind_str[key] = value
                else:
                    param_bind_str[str(key)] = value
            return param_bind_str
        elif isinstance(param, list):
            return [self.__encode(p) for p in param]
        else:
            return param

    def encode(self, o: Any) -> str:
        """
        Return a JSON string representation of a Python data structure.
        """
        new_o = self.__encode(o)
        return super().encode(new_o)

    def default(self, o: Any) -> Any:
        # Convert numpy arrays:
        if hasattr(o, "tolist"):
            return o.tolist()
        # Use Qobj complex json format:
        if isinstance(o, complex):
            return (o.real, o.imag)
        if isinstance(o, ParameterExpression):
            try:
                return float(o)
            except (TypeError, RuntimeError):
                val = complex(o)
                return val.real, val.imag
        return json.JSONEncoder.default(self, o)



# This code is part of Tergite
#
# (C) Copyright Chalmers Next Labs 2024
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import numpy as np
from sympy import And, Piecewise, S, cos, pi, symbols


def delta_t_function(t, args):
    """
    Vectorized envelope function δ(t) as a function of time.

    Returns:
    - delta_t: the value of δ(t) (same shape as t)
    """
    t_w = args["t_w"]
    t_rf = args["t_rf"]
    t_p = args["t_p"]
    delta_0 = args["delta_0"]

    condlist = [
        t <= t_w,
        (t > t_w) & (t <= t_w + t_rf / 2),
        (t > t_w + t_rf / 2) & (t < t_w + t_rf / 2 + t_p),
        (t >= t_w + t_rf / 2 + t_p) & (t < t_w + t_rf + t_p),
        t >= t_w + t_rf + t_p,
    ]
    choicelist = [
        0,
        delta_0 / 2 * (1 - np.cos(2 * np.pi * (t - t_w) / t_rf)),
        delta_0,
        delta_0 / 2 * (1 - np.cos(2 * np.pi * (t - t_w - t_p) / t_rf)),
        0,
    ]
    return np.select(condlist, choicelist)


def delta_t_function_sympy(t, symbolic_args):
    """
    Symbolic envelope function δ(t) as a function of time using SymPy.

    Parameters:
    - t: SymPy symbol representing time.
    - symbolic_args: dictionary containing SymPy symbols for parameters.

    Returns:
    - delta_t: SymPy expression representing δ(t).
    """
    # Unpack symbolic variables, use symbols if not provided
    t_w = symbolic_args.get("t_w", symbols("t_w", real=True))
    t_rf = symbolic_args.get("t_rf", symbols("t_rf", real=True))
    t_p = symbolic_args.get("t_p", symbols("t_p", real=True))
    delta_0 = symbolic_args.get("delta_0", symbols("delta_0", real=True))

    # Define conditions using SymPy relational operators
    cond1 = t <= t_w
    cond2 = And(t > t_w, t <= t_w + t_rf / 2)
    cond3 = And(t > t_w + t_rf / 2, t < t_w + t_rf / 2 + t_p)
    cond4 = And(t >= t_w + t_rf / 2 + t_p, t < t_w + t_rf + t_p)
    cond5 = t >= t_w + t_rf + t_p

    # Define expressions using SymPy functions
    expr1 = S.Zero
    expr2 = delta_0 / 2 * (1 - cos(2 * pi * (t - t_w) / t_rf))
    expr3 = delta_0
    expr4 = delta_0 / 2 * (1 - cos(2 * pi * (t - t_w - t_p) / t_rf))
    expr5 = S.Zero

    # Create the Piecewise function
    delta_t = Piecewise(
        (expr1, cond1),
        (expr2, cond2),
        (expr3, cond3),
        (expr4, cond4),
        (expr5, cond5),
    )
    return delta_t



# This code is part of Tergite
#
# (C) Chalmers Next Labs (2024)
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# ---------------- ALTERATION NOTICE ---------------- #
# This code was modified from the original by:
# Chalmers Next Labs 2024
#
# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#

"""Assemble function for converting a list of circuits into a qobj."""
import hashlib
from collections import defaultdict
from typing import Any, Dict, List, Tuple, Union

from qiskit import pulse
from qiskit.exceptions import QiskitError
from qiskit.pulse import channels, instructions, library, schedule, transforms

from .. import qobj
from ..qobj import converters
from ..qobj import utils as qobj_utils
from ..qobj.converters.pulse_instruction import ParametricPulseShapes
from .run_config import RunConfig


def assemble_schedules(
    schedules: List[
        Union[
            schedule.ScheduleBlock,
            schedule.ScheduleComponent,
            Tuple[int, schedule.ScheduleComponent],
        ]
    ],
    qobj_id: int,
    qobj_header: qobj.QobjHeader,
    run_config: RunConfig,
) -> qobj.PulseQobj:
    """Assembles a list of schedules into a qobj that can be run on the backend.

    Args:
        schedules: Schedules to assemble.
        qobj_id: Identifier for the generated qobj.
        qobj_header: Header to pass to the results.
        run_config: Configuration of the runtime environment.

    Returns:
        The Qobj to be run on the backends.

    Raises:
        QiskitError: when frequency settings are not supplied.

    Examples:

        .. code-block:: python

            from qiskit import pulse
            from qiskit.assembler import assemble_schedules
            from qiskit.assembler.run_config import RunConfig
            # Construct a Qobj header for the output Qobj
            header = {"backend_name": "FakeOpenPulse2Q", "backend_version": "0.0.0"}
            # Build a configuration object for the output Qobj
            config = RunConfig(shots=1024,
                               memory=False,
                               meas_level=1,
                               meas_return='avg',
                               memory_slot_size=100,
                               parametric_pulses=[],
                               init_qubits=True,
                               qubit_lo_freq=[4900000000.0, 5000000000.0],
                               meas_lo_freq=[6500000000.0, 6600000000.0],
                               schedule_los=[])
            # Build a Pulse schedule to assemble into a Qobj
            schedule = pulse.Schedule()
            schedule += pulse.Play(pulse.Waveform([0.1] * 16, name="test0"),
                                   pulse.DriveChannel(0),
                                   name="test1")
            schedule += pulse.Play(pulse.Waveform([0.1] * 16, name="test1"),
                                   pulse.DriveChannel(0),
                                   name="test2")
            schedule += pulse.Play(pulse.Waveform([0.5] * 16, name="test0"),
                                   pulse.DriveChannel(0),
                                   name="test1")
            # Assemble a Qobj from the schedule.
            pulseQobj = assemble_schedules(schedules=[schedule],
                                           qobj_id="custom-id",
                                           qobj_header=header,
                                           run_config=config)
    """
    if not hasattr(run_config, "qubit_lo_freq"):
        raise QiskitError("qubit_lo_freq must be supplied.")
    if not hasattr(run_config, "meas_lo_freq"):
        raise QiskitError("meas_lo_freq must be supplied.")

    lo_converter = converters.LoConfigConverter(
        qobj.PulseQobjExperimentConfig, **run_config.to_dict()
    )
    experiments, experiment_config = _assemble_experiments(
        schedules, lo_converter, run_config
    )
    qobj_config = _assemble_config(lo_converter, experiment_config, run_config)

    return qobj.PulseQobj(
        experiments=experiments, qobj_id=qobj_id, header=qobj_header, config=qobj_config
    )


def _assemble_experiments(
    schedules: List[
        Union[schedule.ScheduleComponent, Tuple[int, schedule.ScheduleComponent]]
    ],
    lo_converter: converters.LoConfigConverter,
    run_config: RunConfig,
) -> Tuple[List[qobj.PulseQobjExperiment], Dict[str, Any]]:
    """Assembles a list of schedules into PulseQobjExperiments, and returns related metadata that
    will be assembled into the Qobj configuration.

    Args:
        schedules: Schedules to assemble.
        lo_converter: The configured frequency converter and validator.
        run_config: Configuration of the runtime environment.

    Returns:
        The list of assembled experiments, and the dictionary of related experiment config.

    Raises:
        QiskitError: when frequency settings are not compatible with the experiments.
    """
    freq_configs = [
        lo_converter(lo_dict) for lo_dict in getattr(run_config, "schedule_los", [])
    ]

    if len(schedules) > 1 and len(freq_configs) not in [0, 1, len(schedules)]:
        raise QiskitError(
            "Invalid 'schedule_los' setting specified. If specified, it should be "
            "either have a single entry to apply the same LOs for each schedule or "
            "have length equal to the number of schedules."
        )

    instruction_converter = getattr(
        run_config, "instruction_converter", converters.InstructionToQobjConverter
    )
    instruction_converter = instruction_converter(
        qobj.PulseQobjInstruction, **run_config.to_dict()
    )

    formatted_schedules = [
        transforms.target_qobj_transform(sched) for sched in schedules
    ]
    compressed_schedules = transforms.compress_pulses(formatted_schedules)

    user_pulselib = {}
    experiments = []
    for idx, sched in enumerate(compressed_schedules):
        qobj_instructions, max_memory_slot = _assemble_instructions(
            sched, instruction_converter, run_config, user_pulselib
        )

        metadata = sched.metadata
        if metadata is None:
            metadata = {}
        # TODO: add other experimental header items (see circuit assembler)
        qobj_experiment_header = qobj.QobjExperimentHeader(
            memory_slots=max_memory_slot + 1,  # Memory slots are 0 indexed
            name=sched.name or f"Experiment-{idx}",
            metadata=metadata,
        )

        experiment = qobj.PulseQobjExperiment(
            header=qobj_experiment_header, instructions=qobj_instructions
        )
        if freq_configs:
            # This handles the cases where one frequency setting applies to all experiments and
            # where each experiment has a different frequency
            freq_idx = idx if len(freq_configs) != 1 else 0
            experiment.config = freq_configs[freq_idx]

        experiments.append(experiment)

    # Frequency sweep
    if freq_configs and len(experiments) == 1:
        experiment = experiments[0]
        experiments = []
        for freq_config in freq_configs:
            experiments.append(
                qobj.PulseQobjExperiment(
                    header=experiment.header,
                    instructions=experiment.instructions,
                    config=freq_config,
                )
            )

    # Top level Qobj configuration
    experiment_config = {
        "pulse_library": [
            qobj.PulseLibraryItem(name=name, samples=samples)
            for name, samples in user_pulselib.items()
        ],
        "memory_slots": max(exp.header.memory_slots for exp in experiments),
    }

    return experiments, experiment_config


def _assemble_instructions(
    sched: Union[pulse.Schedule, pulse.ScheduleBlock],
    instruction_converter: converters.InstructionToQobjConverter,
    run_config: RunConfig,
    user_pulselib: Dict[str, List[complex]],
) -> Tuple[List[qobj.PulseQobjInstruction], int]:
    """Assembles the instructions in a schedule into a list of PulseQobjInstructions and returns
    related metadata that will be assembled into the Qobj configuration. Lookup table for
    pulses defined in all experiments are registered in ``user_pulselib``. This object should be
    mutable python dictionary so that items are properly updated after each instruction assemble.
    The dictionary is not returned to avoid redundancy.

    Args:
        sched: Schedule to assemble.
        instruction_converter: A converter instance which can convert PulseInstructions to
                               PulseQobjInstructions.
        run_config: Configuration of the runtime environment.
        user_pulselib: User pulse library from previous schedule.

    Returns:
        A list of converted instructions, the user pulse library dictionary (from pulse name to
        pulse samples), and the maximum number of readout memory slots used by this Schedule.
    """
    sched = transforms.target_qobj_transform(sched)

    max_memory_slot = 0
    qobj_instructions = []

    acquire_instruction_map = defaultdict(list)
    for time, instruction in sched.instructions:
        if isinstance(instruction, instructions.Play):
            if isinstance(instruction.pulse, library.SymbolicPulse):
                is_backend_supported = True
                try:
                    pulse_shape = ParametricPulseShapes.from_instance(
                        instruction.pulse
                    ).name
                    if pulse_shape not in run_config.parametric_pulses:
                        is_backend_supported = False
                except ValueError:
                    # Custom pulse class, or bare SymbolicPulse object.
                    is_backend_supported = False

                if not is_backend_supported:
                    instruction = instructions.Play(
                        instruction.pulse.get_waveform(),
                        instruction.channel,
                        name=instruction.name,
                    )

            if isinstance(instruction.pulse, library.Waveform):
                name = hashlib.sha256(instruction.pulse.samples).hexdigest()
                instruction = instructions.Play(
                    library.Waveform(name=name, samples=instruction.pulse.samples),
                    channel=instruction.channel,
                    name=name,
                )
                user_pulselib[name] = instruction.pulse.samples

        # ignore explicit delay instrs on acq channels as they are invalid on IBMQ backends;
        # timing of other instrs will still be shifted appropriately
        if isinstance(instruction, instructions.Delay) and isinstance(
            instruction.channel, channels.AcquireChannel
        ):
            continue

        if isinstance(instruction, instructions.Acquire):
            if instruction.mem_slot:
                max_memory_slot = max(max_memory_slot, instruction.mem_slot.index)
            # Acquires have a single AcquireChannel per inst, but we have to bundle them
            # together into the Qobj as one instruction with many channels
            acquire_instruction_map[(time, instruction.duration)].append(instruction)
            continue

        qobj_instructions.append(instruction_converter(time, instruction))

    if acquire_instruction_map:
        if hasattr(run_config, "meas_map"):
            _validate_meas_map(acquire_instruction_map, run_config.meas_map)
        for (time, _), instruction_bundle in acquire_instruction_map.items():
            qobj_instructions.append(
                instruction_converter(time, instruction_bundle),
            )

    return qobj_instructions, max_memory_slot


def _validate_meas_map(
    instruction_map: Dict[Tuple[int, instructions.Acquire], List[instructions.Acquire]],
    meas_map: List[List[int]],
) -> None:
    """Validate all qubits tied in ``meas_map`` are to be acquired.

    Args:
        instruction_map: A dictionary grouping Acquire instructions according to their start time
                         and duration.
        meas_map: List of groups of qubits that must be acquired together.

    Raises:
        QiskitError: If the instructions do not satisfy the measurement map.
    """
    sorted_inst_map = sorted(instruction_map.items(), key=lambda item: item[0])
    meas_map_sets = [set(m) for m in meas_map]

    # error if there is time overlap between qubits in the same meas_map
    for idx, inst in enumerate(sorted_inst_map[:-1]):
        inst_end_time = inst[0][0] + inst[0][1]
        next_inst = sorted_inst_map[idx + 1]
        next_inst_time = next_inst[0][0]
        if next_inst_time < inst_end_time:
            inst_qubits = {inst.channel.index for inst in inst[1]}
            next_inst_qubits = {inst.channel.index for inst in next_inst[1]}
            for meas_set in meas_map_sets:
                common_instr_qubits = inst_qubits.intersection(meas_set)
                common_next = next_inst_qubits.intersection(meas_set)
                if common_instr_qubits and common_next:
                    raise QiskitError(
                        f"Qubits {common_instr_qubits} and {common_next} are in the same measurement "
                        f"grouping: {meas_map}. "
                        "They must either be acquired at the same time, or disjointly"
                        f". Instead, they were acquired at times: {inst[0][0]}-{inst_end_time} and "
                        f"{next_inst_time}-{next_inst_time + next_inst[0][1]}"
                    )


def _assemble_config(
    lo_converter: converters.LoConfigConverter,
    experiment_config: Dict[str, Any],
    run_config: RunConfig,
) -> qobj.PulseQobjConfig:
    """Assembles the QobjConfiguration from experimental config and runtime config.

    Args:
        lo_converter: The configured frequency converter and validator.
        experiment_config: Schedules to assemble.
        run_config: Configuration of the runtime environment.

    Returns:
        The assembled PulseQobjConfig.
    """
    qobj_config = run_config.to_dict()
    qobj_config.update(experiment_config)

    # Run config not needed in qobj config
    qobj_config.pop("meas_map", None)
    qobj_config.pop("qubit_lo_range", None)
    qobj_config.pop("meas_lo_range", None)

    # convert enums to serialized values
    meas_return = qobj_config.get("meas_return", "avg")
    if isinstance(meas_return, qobj_utils.MeasReturnType):
        qobj_config["meas_return"] = meas_return.value

    meas_level = qobj_config.get("meas_level", 2)
    if isinstance(meas_level, qobj_utils.MeasLevel):
        qobj_config["meas_level"] = meas_level.value

    # convert LO frequencies to GHz
    qobj_config["qubit_lo_freq"] = [freq / 1e9 for freq in qobj_config["qubit_lo_freq"]]
    qobj_config["meas_lo_freq"] = [freq / 1e9 for freq in qobj_config["meas_lo_freq"]]

    # override defaults if single entry for ``schedule_los``
    schedule_los = qobj_config.pop("schedule_los", [])
    if len(schedule_los) == 1:
        lo_dict = schedule_los[0]
        q_los = lo_converter.get_qubit_los(lo_dict)
        # Hz -> GHz
        if q_los:
            qobj_config["qubit_lo_freq"] = [freq / 1e9 for freq in q_los]
        m_los = lo_converter.get_meas_los(lo_dict)
        if m_los:
            qobj_config["meas_lo_freq"] = [freq / 1e9 for freq in m_los]

    return qobj.PulseQobjConfig(**qobj_config)



# This code is part of Tergite
#
# (C) Chalmers Next Labs (2024)
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# ---------------- ALTERATION NOTICE ---------------- #
# This code was modified from the original by:
# Chalmers Next Labs 2024
#
# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#


"""Assemble function for converting a list of circuits into a qobj."""
import copy
import warnings
from collections import defaultdict
from typing import Dict, List, Optional, Tuple

from qiskit.circuit import QuantumCircuit
from qiskit.circuit.classicalregister import Clbit
from qiskit.exceptions import QiskitError
from qiskit.utils import deprecate_func
from qiskit.utils.parallel import parallel_map

from ..qobj import (
    GateCalibration,
    PulseLibraryItem,
    PulseQobjInstruction,
    QasmExperimentCalibrations,
    QasmQobj,
    QasmQobjConfig,
    QasmQobjExperiment,
    QasmQobjExperimentConfig,
    QasmQobjInstruction,
    QobjExperimentHeader,
    QobjHeader,
    converters,
)
from .assemble_schedules import _assemble_instructions as _assemble_schedule
from .run_config import RunConfig

PulseLibrary = Dict[str, List[complex]]


def _assemble_circuit(
    circuit: QuantumCircuit, run_config: RunConfig
) -> Tuple[QasmQobjExperiment, Optional[PulseLibrary]]:
    """Assemble one circuit.

    Args:
        circuit: circuit to assemble
        run_config: configuration of the runtime environment

    Returns:
        One experiment for the QasmQobj, and pulse library for pulse gates (which could be None)

    Raises:
        QiskitError: when the circuit has unit other than 'dt'.
    """
    if circuit.unit != "dt":
        raise QiskitError(
            f"Unable to assemble circuit with unit '{circuit.unit}', which must be 'dt'."
        )

    # header data
    num_qubits = 0
    memory_slots = 0
    qubit_labels = []
    clbit_labels = []

    qreg_sizes = []
    creg_sizes = []
    for qreg in circuit.qregs:
        qreg_sizes.append([qreg.name, qreg.size])
        for j in range(qreg.size):
            qubit_labels.append([qreg.name, j])
        num_qubits += qreg.size
    for creg in circuit.cregs:
        creg_sizes.append([creg.name, creg.size])
        for j in range(creg.size):
            clbit_labels.append([creg.name, j])
        memory_slots += creg.size

    qubit_indices = {qubit: idx for idx, qubit in enumerate(circuit.qubits)}
    clbit_indices = {clbit: idx for idx, clbit in enumerate(circuit.clbits)}

    # TODO: why do we need creq_sizes and qreg_sizes in header
    # TODO: we need to rethink memory_slots as they are tied to classical bit
    metadata = circuit.metadata
    if metadata is None:
        metadata = {}
    with warnings.catch_warnings():
        # The class QobjExperimentHeader is deprecated
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
        header = QobjExperimentHeader(
            qubit_labels=qubit_labels,
            n_qubits=num_qubits,
            qreg_sizes=qreg_sizes,
            clbit_labels=clbit_labels,
            memory_slots=memory_slots,
            creg_sizes=creg_sizes,
            name=circuit.name,
            global_phase=float(circuit.global_phase),
            metadata=metadata,
        )

    # TODO: why do we need n_qubits and memory_slots in both the header and the config
    with warnings.catch_warnings():
        # The class QasmQobjExperimentConfig is deprecated
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
        config = QasmQobjExperimentConfig(
            n_qubits=num_qubits, memory_slots=memory_slots
        )
    calibrations, pulse_library = _assemble_pulse_gates(circuit, run_config)
    if calibrations:
        config.calibrations = calibrations

    # Convert conditionals from OpenQASM-2-style (creg ?= int) to qobj-style
    # (register_bit ?= 1), by assuming device has unlimited register slots
    # (supported only for simulators). Map all measures to a register matching
    # their clbit_index, create a new register slot for every conditional gate
    # and add a bfunc to map the creg=val mask onto the gating register bit.

    is_conditional_experiment = any(
        getattr(instruction.operation, "condition", None)
        for instruction in circuit.data
    )
    max_conditional_idx = 0

    instructions = []
    for op_context in circuit.data:
        instruction = op_context.operation._assemble()

        # Add register attributes to the instruction
        qargs = op_context.qubits
        cargs = op_context.clbits
        if qargs:
            instruction.qubits = [qubit_indices[qubit] for qubit in qargs]
        if cargs:
            instruction.memory = [clbit_indices[clbit] for clbit in cargs]
            # If the experiment has conditional instructions, assume every
            # measurement result may be needed for a conditional gate.
            if instruction.name == "measure" and is_conditional_experiment:
                instruction.register = [clbit_indices[clbit] for clbit in cargs]

        # To convert to a qobj-style conditional, insert a bfunc prior
        # to the conditional instruction to map the creg ?= val condition
        # onto a gating register bit.
        if hasattr(instruction, "_condition"):
            ctrl_reg, ctrl_val = instruction._condition
            mask = 0
            val = 0
            if isinstance(ctrl_reg, Clbit):
                mask = 1 << clbit_indices[ctrl_reg]
                val = (ctrl_val & 1) << clbit_indices[ctrl_reg]
            else:
                for clbit in clbit_indices:
                    if clbit in ctrl_reg:
                        mask |= 1 << clbit_indices[clbit]
                        val |= (
                            (ctrl_val >> list(ctrl_reg).index(clbit)) & 1
                        ) << clbit_indices[clbit]

            conditional_reg_idx = memory_slots + max_conditional_idx
            with warnings.catch_warnings():
                # The class QasmQobjInstruction is deprecated
                warnings.filterwarnings(
                    "ignore", category=DeprecationWarning, module="qiskit"
                )
                conversion_bfunc = QasmQobjInstruction(
                    name="bfunc",
                    mask="0x%X" % mask,  # pylint: disable=consider-using-f-string
                    relation="==",
                    val="0x%X" % val,  # pylint: disable=consider-using-f-string
                    register=conditional_reg_idx,
                )
            instructions.append(conversion_bfunc)
            instruction.conditional = conditional_reg_idx
            max_conditional_idx += 1
            # Delete condition attribute now that we have replaced it with
            # the conditional and bfunc
            del instruction._condition

        instructions.append(instruction)
    with warnings.catch_warnings():
        # The class QasmQobjExperiment is deprecated
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
        return (
            QasmQobjExperiment(instructions=instructions, header=header, config=config),
            pulse_library,
        )


def _assemble_pulse_gates(
    circuit: QuantumCircuit, run_config: RunConfig
) -> Tuple[Optional[QasmExperimentCalibrations], Optional[PulseLibrary]]:
    """Assemble and return the circuit calibrations and associated pulse library, if there are any.
    The calibrations themselves may reference the pulse library which is returned as a dict.

    Args:
        circuit: circuit which may have pulse calibrations
        run_config: configuration of the runtime environment

    Returns:
        The calibrations and pulse library, if there are any
    """
    if not circuit.calibrations:
        return None, None
    if not hasattr(run_config, "parametric_pulses"):
        run_config.parametric_pulses = []
    calibrations = []
    pulse_library = {}
    for gate, cals in circuit.calibrations.items():
        for (qubits, params), schedule in cals.items():
            qobj_instructions, _ = _assemble_schedule(
                schedule,
                converters.InstructionToQobjConverter(PulseQobjInstruction),
                run_config,
                pulse_library,
            )
            calibrations.append(
                GateCalibration(
                    str(gate), list(qubits), list(params), qobj_instructions
                )
            )
    return QasmExperimentCalibrations(gates=calibrations), pulse_library


def _extract_common_calibrations(
    experiments: List[QasmQobjExperiment],
) -> Tuple[List[QasmQobjExperiment], Optional[QasmExperimentCalibrations]]:
    """Given a list of ``QasmQobjExperiment``s, each of which may have calibrations in their
    ``config``, collect common calibrations into a global ``QasmExperimentCalibrations``
    and delete them from their local experiments.

    Args:
        experiments: The list of OpenQASM experiments that are being assembled into one qobj

    Returns:
        The input experiments with modified calibrations, and common calibrations, if there
        are any
    """

    def index_calibrations() -> Dict[int, List[Tuple[int, GateCalibration]]]:
        """Map each calibration to all experiments that contain it."""
        exp_indices = defaultdict(list)
        for exp_idx, exp in enumerate(experiments):
            for gate_cal in exp.config.calibrations.gates:
                # They must be keyed on the hash or identical cals will be indexed separately
                exp_indices[hash(gate_cal)].append((exp_idx, gate_cal))
        return exp_indices

    def collect_common_calibrations() -> List[GateCalibration]:
        """If a gate calibration appears in all experiments, collect it."""
        common_calibrations = []
        for _, exps_w_cal in exp_indices.items():
            if len(exps_w_cal) == len(experiments):
                _, gate_cal = exps_w_cal[0]
                common_calibrations.append(gate_cal)
        return common_calibrations

    def remove_common_gate_calibrations(exps: List[QasmQobjExperiment]) -> None:
        """For calibrations that appear in all experiments, remove them from the individual
        experiment's ``config.calibrations``."""
        for _, exps_w_cal in exp_indices.items():
            if len(exps_w_cal) == len(exps):
                for exp_idx, gate_cal in exps_w_cal:
                    exps[exp_idx].config.calibrations.gates.remove(gate_cal)

    if not (
        experiments and all(hasattr(exp.config, "calibrations") for exp in experiments)
    ):
        # No common calibrations
        return experiments, None

    exp_indices = index_calibrations()
    common_calibrations = collect_common_calibrations()
    remove_common_gate_calibrations(experiments)

    # Remove the ``calibrations`` attribute if it's now empty
    for exp in experiments:
        if not exp.config.calibrations.gates:
            del exp.config.calibrations

    return experiments, QasmExperimentCalibrations(gates=common_calibrations)


def _configure_experiment_los(
    experiments: List[QasmQobjExperiment],
    lo_converter: converters.LoConfigConverter,
    run_config: RunConfig,
):
    # get per experiment los
    freq_configs = [
        lo_converter(lo_dict) for lo_dict in getattr(run_config, "schedule_los", [])
    ]

    if len(experiments) > 1 and len(freq_configs) not in [0, 1, len(experiments)]:
        raise QiskitError(
            "Invalid 'schedule_los' setting specified. If specified, it should be "
            "either have a single entry to apply the same LOs for each experiment or "
            "have length equal to the number of experiments."
        )

    if len(freq_configs) > 1:
        if len(experiments) > 1:
            for idx, expt in enumerate(experiments):
                freq_config = freq_configs[idx]
                expt.config.qubit_lo_freq = freq_config.qubit_lo_freq
                expt.config.meas_lo_freq = freq_config.meas_lo_freq
        elif len(experiments) == 1:
            expt = experiments[0]
            experiments = []
            for freq_config in freq_configs:
                expt_config = copy.deepcopy(expt.config)
                expt_config.qubit_lo_freq = freq_config.qubit_lo_freq
                expt_config.meas_lo_freq = freq_config.meas_lo_freq
                experiments.append(
                    QasmQobjExperiment(
                        header=expt.header,
                        instructions=expt.instructions,
                        config=expt_config,
                    )
                )

    return experiments


def _assemble_circuits(
    circuits: List[QuantumCircuit],
    run_config: RunConfig,
    qobj_id: int,
    qobj_header: QobjHeader,
) -> QasmQobj:
    with warnings.catch_warnings():
        # Still constructs Qobj, that is deprecated. The message is hard to trace to a module,
        # because concurrency is hard.
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        experiments_and_pulse_libs = parallel_map(
            _assemble_circuit, circuits, [run_config]
        )
    experiments = []
    pulse_library = {}
    for exp, lib in experiments_and_pulse_libs:
        experiments.append(exp)
        if lib:
            pulse_library.update(lib)

    # extract common calibrations
    experiments, calibrations = _extract_common_calibrations(experiments)

    # configure LO freqs per circuit
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
        lo_converter = converters.LoConfigConverter(
            QasmQobjExperimentConfig, **run_config.to_dict()
        )
    experiments = _configure_experiment_los(experiments, lo_converter, run_config)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
        qobj_config = QasmQobjConfig()
    if run_config:
        qobj_config_dict = run_config.to_dict()

        # remove LO ranges, not needed in qobj
        qobj_config_dict.pop("qubit_lo_range", None)
        qobj_config_dict.pop("meas_lo_range", None)

        # convert LO frequencies to GHz, if they exist
        if "qubit_lo_freq" in qobj_config_dict:
            qobj_config_dict["qubit_lo_freq"] = [
                freq / 1e9 for freq in qobj_config_dict["qubit_lo_freq"]
            ]
        if "meas_lo_freq" in qobj_config_dict:
            qobj_config_dict["meas_lo_freq"] = [
                freq / 1e9 for freq in qobj_config_dict["meas_lo_freq"]
            ]

        # override default los if single ``schedule_los`` entry set
        schedule_los = qobj_config_dict.pop("schedule_los", [])
        if len(schedule_los) == 1:
            lo_dict = schedule_los[0]
            q_los = lo_converter.get_qubit_los(lo_dict)
            # Hz -> GHz
            if q_los:
                qobj_config_dict["qubit_lo_freq"] = [freq / 1e9 for freq in q_los]
            m_los = lo_converter.get_meas_los(lo_dict)
            if m_los:
                qobj_config_dict["meas_lo_freq"] = [freq / 1e9 for freq in m_los]

        with warnings.catch_warnings():
            # The class QasmQobjConfig is deprecated
            warnings.filterwarnings(
                "ignore", category=DeprecationWarning, module="qiskit"
            )
            qobj_config = QasmQobjConfig(**qobj_config_dict)

    qubit_sizes = []
    memory_slot_sizes = []
    for circ in circuits:
        num_qubits = 0
        memory_slots = 0
        for qreg in circ.qregs:
            num_qubits += qreg.size
        for creg in circ.cregs:
            memory_slots += creg.size
        qubit_sizes.append(num_qubits)
        memory_slot_sizes.append(memory_slots)
    qobj_config.memory_slots = max(memory_slot_sizes)
    qobj_config.n_qubits = max(qubit_sizes)

    if pulse_library:
        qobj_config.pulse_library = [
            PulseLibraryItem(name=name, samples=samples)
            for name, samples in pulse_library.items()
        ]

    if calibrations and calibrations.gates:
        qobj_config.calibrations = calibrations
    with warnings.catch_warnings():
        # The class QasmQobj is deprecated
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
        return QasmQobj(
            qobj_id=qobj_id,
            config=qobj_config,
            experiments=experiments,
            header=qobj_header,
        )


@deprecate_func(
    since="1.2",
    removal_timeline="in the 2.0 release",
    additional_msg="The `Qobj` class and related functionality are part of the deprecated `BackendV1` "
    "workflow,  and no longer necessary for `BackendV2`. If a user workflow requires "
    "`Qobj` it likely relies on deprecated functionality and should be updated to "
    "use `BackendV2`.",
)
def assemble_circuits(
    circuits: List[QuantumCircuit],
    run_config: RunConfig,
    qobj_id: int,
    qobj_header: QobjHeader,
) -> QasmQobj:
    """Assembles a list of circuits into a qobj that can be run on the backend.

    Args:
        circuits: circuit(s) to assemble
        run_config: configuration of the runtime environment
        qobj_id: identifier for the generated qobj
        qobj_header: header to pass to the results

    Returns:
        The qobj to be run on the backends

    Examples:

        .. code-block:: python

            from qiskit.circuit import QuantumRegister, ClassicalRegister, QuantumCircuit
            from qiskit.assembler import assemble_circuits
            from qiskit.assembler.run_config import RunConfig
            # Build a circuit to convert into a Qobj
            q = QuantumRegister(2)
            c = ClassicalRegister(2)
            qc = QuantumCircuit(q, c)
            qc.h(q[0])
            qc.cx(q[0], q[1])
            qc.measure(q, c)
            # Assemble a Qobj from the input circuit
            qobj = assemble_circuits(circuits=[qc],
                                     qobj_id="custom-id",
                                     qobj_header=[],
                                     run_config=RunConfig(shots=2000, memory=True, init_qubits=True))
    """
    # assemble the circuit experiments
    return _assemble_circuits(circuits, run_config, qobj_id, qobj_header)



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Alteration Notice
# -----------------
# This code was refactored from the original by:
#
# Martin Ahindura, 2023
"""Handles the compilation of device properties into qiskit target"""
from typing import TYPE_CHECKING

import qiskit.circuit as circuit
from qiskit.transpiler import InstructionProperties, Target

from ..device_compiler import schedules as templates

if TYPE_CHECKING:
    from ...types.backend import TergiteBackend
    from ..api_client.dtos import DeviceCalibration


def add_instructions(
    target: Target,
    backend: "TergiteBackend",
    qubits: tuple,
    coupled_qubit_idxs: tuple,
    device_properties: "DeviceCalibration",
) -> Target:
    """Adds Rx, Rz gates and Delay and Measure instructions for each qubit
    to the transpiler target provided.

    Implementations are backend dependent and are based on a specific calibration.

    Args:
        target: the target on which the instruction is to be added
        backend: the instance of
            :class:`~tergite.providers.tergite.backend:TergiteBackend`
            for which to add the instructions
        qubits: a tuple of qubits on which the instruction is to be run
        coupled_qubit_idxs: a tuple of tuples of qubit indexes that are coupled
        device_properties: the dynamic device parameters of the device
    """
    # TODO: Fetch error statistics of gates from database

    rx_theta = circuit.Parameter("theta")
    # changed lambda to lambda_param as it's not a valid name when it's used as a parameter further in transpilation
    rz_lambda = circuit.Parameter("lambda_param")
    delay_tau = circuit.Parameter("tau")

    # Reset qubits to ground state
    reset_props = {
        (q,): InstructionProperties(
            error=0.0,
            calibration=templates.delay(backend, (q,), 300 * 1000, delay_str="Reset"),
        )
        for q in qubits
    }
    target.add_instruction(circuit.library.Reset(), reset_props)

    # Rotation around X-axis on Bloch sphere
    rx_props = {
        (q,): InstructionProperties(
            error=0.0,
            calibration=templates.rx(
                backend, (q,), rx_theta, device_properties=device_properties
            ),
        )
        for q in qubits
    }
    target.add_instruction(circuit.library.standard_gates.RXGate(rx_theta), rx_props)

    # Rotation around Z-axis on Bloch sphere
    rz_props = {
        (q,): InstructionProperties(
            error=0.0, calibration=templates.rz(backend, (q,), rz_lambda)
        )
        for q in qubits
    }
    target.add_instruction(circuit.library.standard_gates.RZGate(rz_lambda), rz_props)

    # Delay instruction
    delay_props = {
        (q,): InstructionProperties(
            error=0.0, calibration=templates.delay(backend, (q,), delay_tau)
        )
        for q in qubits
    }
    target.add_instruction(circuit.Delay(delay_tau), delay_props)

    measure_instruction = circuit.measure.Measure()

    # Measurement

    # Add all the qubits that are going to be measured
    measure_props = {
        (q,): InstructionProperties(
            error=0.0,
            calibration=templates.measure(
                backend, [q], device_properties=device_properties
            ),
        )
        for q in qubits
    }
    target.add_instruction(measure_instruction, measure_props)

    if coupled_qubit_idxs:
        cz_props = {
            (control_qubit_idx, target_qubit_idx): InstructionProperties(
                error=0.0,
                calibration=templates.cz(
                    backend,
                    control_qubit_idxs=[control_qubit_idx],
                    target_qubit_idxs=[target_qubit_idx],
                    device_properties=device_properties,
                ),
            )
            for control_qubit_idx, target_qubit_idx in coupled_qubit_idxs
        }

        target.add_instruction(circuit.library.CZGate(), cz_props)

    return target



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
# (C) Copyright Chalmers Next Labs 2025
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Utility functions related to the job file"""
import json
from pathlib import Path
from typing import Any, Dict

import h5py

from tergite.compat.qiskit.qobj import PulseQobj


def extract_job_metadata(logfile: Path) -> Dict[str, Any]:
    """Extracts the metadata of the job from the HDF5 logfile

    It converts any numpy types to python types

    Args:
        logfile: the path to the logfile

    Returns:
        the metadata extracted from the file at the logfile path

    Raises:
        RuntimeError: malformed logfile. '{str}' not found
    """
    with h5py.File(logfile, "r") as hdf:
        try:
            return {
                k: _np_to_py(v) for k, v in hdf["header/qobj_metadata"].attrs.items()
            }
        except KeyError as exp:
            raise RuntimeError(f"malformed logfile. '{exp}' not found")


def extract_job_qobj(logfile: Path) -> PulseQobj:
    """Extracts the PulseQobj from the HDF5 logfile

    Args:
        logfile: the path to the logfile

    Returns:
        the PulseQobj derived from the file at the logfile path

    Raises:
        RuntimeError: malformed logfile. '{str}' not found
    """
    with h5py.File(logfile, "r") as hdf:
        try:
            experiment_data = hdf["header/qobj_data"].attrs.get("experiment_data")
            return PulseQobj.from_dict(data=json.loads(experiment_data))
        except KeyError as exp:
            raise RuntimeError(f"malformed logfile. '{exp}' not found")


def _np_to_py(value: Any) -> Any:
    """Converts a potentially numpy typed value to a python inbuilt type

    Args:
        value: the value to convert

    Returns:
        the value with its python builtin type
    """
    try:
        return value.item()
    except AttributeError:
        return value



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Alteration Notice
# -----------------
# This code was refactored from the original by:
#
# Martin Ahindura, 2023
"""Defines the asynchronous job that executes the experiments."""
import logging
from collections import Counter
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import requests
from qiskit.providers import JobV1
from qiskit.providers.jobstatus import JOB_FINAL_STATES, JobStatus
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

from ..compat.qiskit.qobj import PulseQobj
from ..services import api_client as client
from ..services.api_client import JobFile
from ..services.api_client.dtos import RemoteJobStatus

# from ..services.api_client.dtos import RemoteJobStatus
from ..utils.qobj import compress_qobj_dict

if TYPE_CHECKING:
    from .backend import TergiteBackend
    from .provider import Provider

_JOB_FINAL_OR_INITIAL_STATES = (*JOB_FINAL_STATES, JobStatus.INITIALIZING)


class Job(JobV1):
    """A representation of the asynchronous job that handles experiments on a backend"""

    def __init__(
        self,
        *,
        backend: "TergiteBackend",
        job_id: str,
        payload: Optional[PulseQobj] = None,
        upload_url: Optional[str] = None,
        remote_data: Optional["client.RemoteJob"] = None,
        logfile: Optional[Path] = None,
        status: JobStatus = JobStatus.INITIALIZING,
        download_url: Optional[str] = None,
        calibration_date: Optional[str] = None,
        access_token: Optional[str] = None,
        **kwargs,
    ):
        """Initializes the job instance for the given backend

        Args:
            backend: the backed where the job is to run
            job_id: the unique id of the job
            upload_url: URL where the jobs will be uploaded
            payload: the qobj representing the experiments
            remote_data: the job data from the remote API
            logfile: the HDF5 logfile for this job
            status: the JobStatus of the current job; defaults to `JobStatus.INITIALIZING`
            download_url: the URL to download the results logfile from
            calibration_date: the last_calibrated timestamp of the backend when this job was compiled
            access_token: the access token for submitting jobs to BCC
            kwargs: extra key-word arguments to add to the job's metadata
        """
        super().__init__(
            backend=backend, job_id=job_id, upload_url=upload_url, **kwargs
        )
        self.payload = payload
        self.upload_url = upload_url
        self._status = status
        self._logfile = logfile
        self._provider: Provider = backend.provider
        self._download_url = download_url
        self._calibration_date = calibration_date
        self._result: Optional[Result] = None
        self._remote_data = remote_data
        self._account = backend.provider.account
        self._access_token = access_token

    @property
    def _is_in_final_state(self):
        """Whether this job has reached the end of the line or not

        It returns True if the job status is in `CANCELLED`, `ERROR`, `DONE`
        """
        try:
            return self._status in JOB_FINAL_STATES
        except RuntimeError:
            return False

    @property
    def remote_data(self) -> Optional["client.RemoteJob"]:
        """The representation of the job in the remote API"""
        if not self._is_in_final_state:
            self._remote_data = client.get_remote_job_data(self._account, self.job_id())

        return self._remote_data

    def status(self) -> JobStatus:
        if not self._is_in_final_state:
            try:
                self._status = STATUS_MAP[self.remote_data.status]
            except (KeyError, AttributeError):
                pass

        return self._status

    def submit(self) -> requests.Response:
        """Submit the job to the backend for execution.

        Returns:
            requests.Response: the response of the API after submitting the job
        """
        if self._status != JobStatus.INITIALIZING:
            raise ValueError("This job was already submitted")

        if self.upload_url is None:
            raise ValueError("This job is not submittable. It lacks an upload_url")

        if not isinstance(self.payload, PulseQobj):
            raise RuntimeError(f"Unprocessable payload type: {type(self.payload)}")

        if self._access_token is None:
            raise ValueError("This job is not submittable. It lacks an access token")

        payload = compress_qobj_dict(self.payload.to_dict())
        job_entry = JobFile.model_validate(
            {
                "job_id": self.job_id(),
                "params": {"qobj": payload},
            }
        )

        return client.send_job_file(
            self._account,
            url=self.upload_url,
            job_data=job_entry,
            access_token=self._access_token,
        )

    @property
    def download_url(self) -> Optional[str]:
        """The download_url of this job when it is completed

        Raises:
            RuntimeError: Failed to GET download URL of job: {job_id}. Status: {JobStatus}
        """
        if self._download_url is None and self.status() in JOB_FINAL_STATES:
            try:
                self._download_url = self.remote_data.download_url
            except KeyError:
                raise RuntimeError(
                    f"Failed to GET download URL of job: {self.job_id()}. Status: {self.status()}"
                )

        return self._download_url

    @property
    def logfile(self) -> Optional[Path]:
        """The path to the logfile of this job when it is completed"""
        if self._logfile is None and self.status() in JOB_FINAL_STATES:
            if self.download_url:
                self._logfile = client.download_job_logfile(
                    self.job_id(),
                    url=self.download_url,
                    access_token=self._access_token,
                )

        return self._logfile

    def cancel(self):
        """Attempt to cancel the job.

        Raises:
            ValueError: This job is not cancellable. It lacks an upload_url
            RuntimeError: Failed to cancel job '{job_id}': {detail}
        """
        if self.upload_url is None:
            raise ValueError("This job is not cancellable. It lacks an upload_url")

        client.cancel_job(
            upload_url=self.upload_url,
            job_id=self.job_id(),
            access_token=self._access_token,
        )

    def result(self) -> Optional[Result]:
        """Retrieves the outcome of this job when it is completed.

        It returns None if the job has not yet completed

        Returns:
            Optional[qiskit.result.result.Result]: the outcome of this job
                if it has completed

        Raises:
            RuntimeError: failed to GET memory of job: {job_id}
            RuntimeError: unexpected number of results; expected {num}, got: {num}
        """
        status = self.status()
        if status not in JOB_FINAL_STATES:
            logging.info(f"Job {self.job_id()} is not DONE. Status: {status}.")
            return

        if self._result is None:
            backend: TergiteBackend = self.backend()
            n_qubits = self.payload.config.n_qubits

            try:
                memory = self.remote_data.result.memory
            except AttributeError:
                raise RuntimeError(f"failed to GET memory of job: {self.job_id()}")

            # Sanity check
            if len(memory) != len(self.payload.experiments):
                raise RuntimeError(
                    f"unexpected number of results;"
                    f"expected {len(self.payload.experiments)}, got: {len(memory)}"
                )
            meas_level = self.payload.config.meas_level
            has_counts = meas_level == 2

            # Prefer the experiment header width, then qobj config, then n_qubits
            default_slots = (
                getattr(self.payload.config, "memory_slots", None) or n_qubits
            )
            for exp in self.payload.experiments:
                if getattr(exp.header, "memory_slots", None) is None:
                    setattr(exp.header, "memory_slots", default_slots)

            self._result = Result(
                backend_name=backend.name,
                backend_version=backend.backend_version,
                qobj_id=self.payload.qobj_id,
                job_id=self.job_id(),
                success=True,
                results=[
                    ExperimentResult(
                        header=self.payload.experiments[idx].header,
                        shots=self.payload.config.shots,
                        success=True,
                        meas_level=meas_level,
                        data=ExperimentResultData(
                            counts=dict(Counter(v)) if has_counts else None,
                            memory=v,
                        ),
                    )
                    for idx, v in enumerate(memory)
                ],
            )

        return self._result

    def __repr__(self):
        kwargs = [f"{k}={repr(v)}" for k, v in self.__dict__.items()]
        kwargs_str = ",\n".join(kwargs)
        return f"{self.__class__.__name__}({kwargs_str})"

    def __eq__(self, other):
        if not isinstance(other, Job):
            return False

        for k, v in self.__dict__.items():
            other_v = getattr(other, k, None)
            if other_v != v:
                return False

        return True


STATUS_MAP = {
    RemoteJobStatus.PENDING: JobStatus.QUEUED,
    RemoteJobStatus.SUCCESSFUL: JobStatus.DONE,
    RemoteJobStatus.EXECUTING: JobStatus.RUNNING,
    RemoteJobStatus.CANCELLED: JobStatus.CANCELLED,
    RemoteJobStatus.FAILED: JobStatus.ERROR,
}



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
# (C) Copyright Chalmers Next Labs 2025
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Utility function for handling hte Qobj for a job"""
from itertools import groupby
from typing import Any, Dict, List, Optional, Tuple, Union


def compress_qobj_dict(qobj_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Creates a compressed dictionary representation of the qobj dict

    In order to reduce the bandwidth taken up be the qobject
    dict, we do a few things with the data which will be reversed
    at the backend

    Note that this compression is in-place

    Args:
        qobj_dict: the dict of the PulseQobj to compress

    Returns:
        A compressed dict of the qobj
    """
    # In-place RLE pulse library for compression
    for pulse in qobj_dict["config"]["pulse_library"]:
        pulse["samples"] = _iqx_rle(pulse["samples"])

    return qobj_dict


def _iqx_rle(seq: List[Any]) -> List[Union[Tuple[Any], Tuple[Any, int]]]:
    """Run-length encodes a sequence.

    Constant subsequences are stored as a single term and count, rather than as the original subsequence.
    Counts equal to 1 are omitted.

    Args:
        seq: the sequence to encode

    Returns:
        A list of tuples that represent the encoding for the sequence
    """
    seq = [(k, sum(1 for _ in g)) for k, g in groupby(seq)]
    return [(c, rep) if rep > 1 else (c,) for c, rep in seq]


def derive_reg_len_from_qobj_experiment(qobj_exp) -> Optional[int]:
    """Fallback reg length from acquire memory slots."""
    mem_slots: List[int] = []
    for inst in qobj_exp.instructions:
        if getattr(inst, "name", None) != "acquire":
            continue
        ms = getattr(inst, "memory_slot", None)
        if ms is None:
            continue
        if isinstance(ms, list):
            mem_slots.extend(ms)
        else:
            mem_slots.append(int(ms))
    return (max(mem_slots) + 1) if mem_slots else None


def rewrite_acquire_memory_slots(qobj_exp, q_to_c: Dict[int, int]) -> None:
    """Rewrite acquire.memory_slot so it matches the circuit clbit indices."""
    for inst in qobj_exp.instructions:
        if getattr(inst, "name", None) != "acquire":
            continue

        qubits = getattr(inst, "qubits", None)
        mem = getattr(inst, "memory_slot", None)
        if qubits is None or mem is None:
            continue

        # Normalize to lists
        q_list = qubits if isinstance(qubits, list) else [int(qubits)]
        m_list = mem if isinstance(mem, list) else [int(mem)]

        if len(q_list) != len(m_list):
            raise ValueError("Acquire has mismatched qubits/memory_slot lengths")

        new_m_list: List[int] = []
        for q in q_list:
            if q not in q_to_c:
                raise ValueError(
                    f"Qobj metadata quantum to classical mapping doesn't have qubit {q} in {q_to_c.keys()}."
                )
            else:
                new_m_list.append(q_to_c[q])

        inst.memory_slot = new_m_list



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
# (C) Copyright Axel Andersson 2022
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Alteration Notice
# -----------------
# This code was refactored from the original by:
#
# Martin Ahindura, 2023, 2025
"""Defines the backend types provided by Tergite."""

from __future__ import annotations

import functools
import logging
import warnings
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import qiskit.circuit as circuit
import qiskit.compiler as compiler
import qiskit.pulse as pulse
from numpy import inf as infinity
from qiskit.circuit import QuantumCircuit
from qiskit.providers import BackendV2, Options
from qiskit.pulse.channels import (
    AcquireChannel,
    DriveChannel,
    MeasureChannel,
    MemorySlot,
)
from qiskit.transpiler import Target
from qiskit.transpiler.coupling import CouplingMap

from ..compat.qiskit.compiler.assembler import assemble
from ..compat.qiskit.qobj import PulseQobj, QasmQobj
from ..compat.qiskit_ibm_runtime.models import BackendConfiguration
from ..services import api_client, device_compiler
from ..utils.qobj import (
    derive_reg_len_from_qobj_experiment,
    rewrite_acquire_memory_slots,
)
from ..utils.quantum_circuit import (
    as_circuit_list,
    extract_q_to_clbit_map,
    normalise_classical_registers,
)
from .job import Job

if TYPE_CHECKING:
    from ..services.api_client import DeviceCalibration, TergiteBackendConfig
    from .provider import Provider as TergiteProvider


class TergiteBackend(BackendV2):
    """Abstract class for Tergite Backends"""

    max_shots = infinity
    max_circuits = infinity
    provider: "TergiteProvider"

    def __init__(
        self,
        /,
        *,
        data: "TergiteBackendConfig",
        provider: "TergiteProvider",
        base_url: str,
    ):
        """Initialize a TergiteBackend based backend

        Args:
            provider: An optional backwards reference to the
                :class:`~qiskit.providers.Provider` object that the backend
                is from
            data: An optional config of type :class:`TergiteBackendConfig`
                from which to construct this backend
            base_url: An optional URL to Tergite API through which
                jobs are to be run

        Raises:
            AttributeError: If a field is specified that's outside the backend's
                options
            TypeError: If ``data`` is not a :class:`TergiteBackendConfig` object
        """
        super().__init__(
            provider=provider,
            name=data.name,
            backend_version=data.version,
        )
        self.base_url = base_url
        self.data = data

    @property
    @abstractmethod
    def device_properties(self) -> Optional["DeviceCalibration"]:
        """The calibrated data for the given device"""

    def register_job(self, payload: Union[PulseQobj, QasmQobj], **metadata) -> Job:
        """Registers a new asynchronous job with the Tergite API.

        Args:
            payload: the experiments to attach to the given job
            metadata: the extra properties to add to the metadata of the object

        Returns:
             tergite.qiskit.providers.job.Job: An asynchronous
                 job registered in the Tergite API to be executed
        """
        calibration_date = metadata.get("calibration_date", None)
        resp = api_client.register_job(
            self.provider.account,
            backend_name=self.name,
            calibration_date=calibration_date,
        )
        return Job(
            backend=self,
            job_id=resp.job_id,
            payload=payload,
            upload_url=resp.upload_url,
            access_token=resp.access_token,
            **metadata,
        )

    def run(self, experiments, /, **kwargs) -> Job:
        """Run on the Tergite backend.

        This method returns a :class:`~qiskit.providers.Job` object
        that runs circuits. Depending on the backend this may be either an async
        or sync call. It is at the discretion of the provider to decide whether
        running should block until the execution is finished or not: the Job
        class can handle either situation.

        Args:
            experiments (QuantumCircuit or Schedule or ScheduleBlock or list): An
                individual or a list of
                :class:`~qiskit.circuits.QuantumCircuit,
                :class:`~qiskit.pulse.ScheduleBlock`, or
                :class:`~qiskit.pulse.Schedule` objects to run on the backend.
            kwargs: Any kwarg options to pass to the backend for running the
                config. If a key is also present in the options
                attribute/object then the expectation is that the value
                specified will be used instead of what's set in the options
                object.

        Returns:
            tergite.qiskit.providers.job.Job: The job object for the run
        """
        experiments = as_circuit_list(experiments)

        # only normalize if that's a circuit
        if all(isinstance(exp, QuantumCircuit) for exp in experiments):
            # measure_all creates an additional classical register with a different name
            # remove unsused classical register if any present
            experiments = [
                normalise_classical_registers(exp, prefer_c=True) for exp in experiments
            ]

        qobj = self.make_qobj(experiments, **kwargs)
        job = self.register_job(
            payload=qobj,
            shots=qobj.config.shots,
            qobj_id=qobj.qobj_id,
            num_experiments=len(qobj.experiments),
            calibration_date=self.last_calibrated,
        )
        response = job.submit()
        if response.ok:
            print("Tergite: Job has been successfully submitted")
            return job
        else:
            raise RuntimeError(
                f"Unable to transmit job to the Tergite BCC, response: {response}"
            )

    @abstractmethod
    def configuration(self) -> BackendConfiguration:
        """Retrieves this backend's configuration

        Returns:
            qiskit.providers.models.backendconfiguration.BackendConfiguration:
                this backend's configuration
        """
        ...

    @abstractmethod
    def make_qobj(self, experiments: list, /, **kwargs) -> Union[PulseQobj, QasmQobj]:
        """Constructs a Qobj from a list of user OpenPulse schedules or OpenQASM circuits

        Args:
            experiments (QuantumCircuit or Schedule or ScheduleBlock or list): An
                individual or a list of
                :class:`~qiskit.circuits.QuantumCircuit,
                :class:`~qiskit.pulse.ScheduleBlock`, or
                :class:`~qiskit.pulse.Schedule` objects to run on the backend.

        Returns:
             qiskit.qobj.pulse_qobj.PulseQobj or qiskit.qobj.pulse_qobj.QasmQobj
                transpiled from the experiments.
        """
        ...

    @property
    def meas_map(self) -> list:
        return self.data.meas_map

    @property
    def last_calibrated(self) -> Optional[str]:
        """The timestamp for when this backed was last calibrated"""
        try:
            return self.device_properties.last_calibrated
        except AttributeError:
            pass

    @functools.cached_property
    def coupling_map(self) -> CouplingMap:
        # It is required that nodes are contiguously indexed starting at 0.
        # Missed nodes will be added as isolated nodes in the coupling map.
        #
        # Also for some reason the graph has to be connected in Qiskit? Not sure why that is..
        edges = sorted(
            sorted(self.data.coupling_map, key=lambda i: i[1]), key=lambda i: i[0]
        )
        return CouplingMap(couplinglist=list(edges))

    @classmethod
    def _default_options(cls, /) -> Options:
        """This defines the default user configurable settings of this backend.

        See: help(backend.set_options)

         Returns:
            qiskit.providers.Options: A options object with
                default values set
        """
        options = Options(shots=2000)
        options.set_validator("shots", (1, TergiteBackend.max_shots))
        return options

    def _as_dict(self):
        """Converts this backend into a dictionary representation

        Returns:
            dict: a dictionary representation of this backend
        """
        obj = self.configuration().to_dict()
        obj["characterized"] = self.data.characterized
        return obj

    def __repr__(self) -> str:
        repr_list = [f"TergiteBackend object @ {hex(id(self))}:"]
        config = self._as_dict()
        for attr, value in config.items():
            repr_list.append(f"  {attr}:\t{value}".expandtabs(30))
        return "\n".join(repr_list)

    @property
    def backend_id(self) -> tuple:
        # fields that uniquely identify the remote backend
        # FIXME: backend_name is unresolved reference.
        return self.backend_name, self.backend_version, self.base_url

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TergiteBackend):
            return False

        self_dict = self._as_dict().copy()
        other_dict = other._as_dict().copy()

        # serialize a few items that are hard to serialize
        self_dict["coupling_map"] = f"{self_dict['coupling_map']}"
        other_dict["coupling_map"] = f"{other_dict['coupling_map']}"

        self_dict["supported_instructions"] = f"{self_dict['supported_instructions']}"
        other_dict["supported_instructions"] = f"{other_dict['supported_instructions']}"

        return self_dict == other_dict

    def __hash__(self) -> int:
        return hash(self.backend_id)


class OpenPulseBackend(TergiteBackend):
    open_pulse = True
    provider: "TergiteProvider"
    parametric_pulses = [
        "constant",
        "zero",
        "square",
        "sawtooth",
        "triangle",
        "cos",
        "sin",
        "gaussian",
        "gaussian_deriv",
        "sech",
        "sech_deriv",
        "gaussian_square",
        "drag",
        "wacqt_cz_gate_pulse",
    ]

    def __init__(
        self,
        /,
        *,
        data: "TergiteBackendConfig",
        provider: "TergiteProvider",
        base_url: str,
    ):
        self._target: Optional[Target] = None
        self._device_properties: Optional[DeviceCalibration] = None
        super().__init__(data=data, provider=provider, base_url=base_url)

    def configuration(self) -> BackendConfiguration:
        return BackendConfiguration(
            backend_name=self.name,  # From BackendV2.
            backend_version=self.backend_version,  # From BackendV2.
            n_qubits=self.target.num_qubits,  # Number of qubits, obtained from self.target.
            basis_gates=[],  # Don't need.
            gates=[],  # Don't need.
            simulator=False,  # This is a real quantum computer.
            conditional=False,  # We cannot do conditional gate application yet.
            local=False,  # Jobs are sent over the internet.
            open_pulse=True,  # This backend converts all quantum experiment input to schedules.
            meas_levels=(0, 1, 2),  # 0: RAW, 1: INTEGRATED, 2: DISCRIMINATED
            memory=False,  # Unsure what this is for. (?)
            max_shots=TergiteBackend.max_shots,  # From TergiteBackend.
            coupling_map=self.coupling_map,  # From TergiteBackend.
            supported_instructions=self.target.instructions,  # Supported instructions, obtained from self.target.
            dt=self.dt,  # From self
            dtm=self.dtm,  # From self
            description=self.description,  # From BackendV2.
            parametric_pulses=OpenPulseBackend.parametric_pulses,  # From type of self
        )

    @property
    def device_properties(self) -> DeviceCalibration:
        """The calibrated data for the given device"""
        if self._device_properties is None:
            self._refresh_device_properties()
        return self._device_properties

    @property
    def dt(self) -> float:
        return self.data.dt

    @property
    def dtm(self) -> float:
        return self.data.dtm

    @property
    def target(self) -> Target:
        """A qiskit. transpiler. Target object for the backend.

        This internally makes a call to the calibrations endpoint
        to get the latest calibration data and compiles a new Target
        for it.
        """
        if self._target is None:
            self._refresh_target()

        return self._target

    @property
    def qubit_lo_freq(self) -> list:
        # FIXME: I think self.data might have a 'qubit_lo_freq' value.
        return [0.0] * self.data.number_of_qubits

    @property
    def meas_lo_freq(self) -> list:
        # FIXME: I think self.data might have a 'meas_lo_freq' value.
        return [0.0] * self.data.number_of_resonators

    def drive_channel(self, qubit_idx: int) -> DriveChannel:
        return DriveChannel(qubit_idx)

    def measure_channel(self, qubit_idx: int) -> MeasureChannel:
        return MeasureChannel(qubit_idx)

    def acquire_channel(self, qubit_idx: int) -> AcquireChannel:
        return AcquireChannel(qubit_idx)

    def memory_slot(self, qubit_idx: int) -> MemorySlot:
        return MemorySlot(qubit_idx)

    def control_channel(self, qubits: Tuple[int, int]):
        """Return the control channel for the given qubits."""
        try:
            pulse_channel = self.data.qubit_ids_coupler_dict[qubits]
            return [pulse.ControlChannel(pulse_channel)]
        except KeyError:
            raise ValueError(f"Coupling {qubits} not in coupling map.")

    def make_qobj(self, experiments: object, /, **kwargs) -> PulseQobj:
        """Constructs a Qobj from a list of user OpenPulse schedules or OpenQASM circuits

        Args:
            experiments (QuantumCircuit or Schedule or ScheduleBlock or list): An
                individual or a list of
                :class:`~qiskit.circuits.QuantumCircuit,
                :class:`~qiskit.pulse.ScheduleBlock`, or
                :class:`~qiskit.pulse.Schedule` objects to run on the backend.

        Returns:
             qiskit.qobj.pulse_qobj.PulseQobj or qiskit.qobj.pulse_qobj.QasmQobj
                transpiled from the experiments.
        """
        # Ensure that the target is recompiled for the latest calibration data
        self._refresh_target()

        if type(experiments) is not list:
            experiments = [experiments]

        # capture circuit-level classical register metadata before converting to pulses
        per_exp_meta: List[Dict[str, Any]] = []
        for exp in experiments:
            if isinstance(exp, QuantumCircuit):
                per_exp_meta.append(
                    {
                        "c_reg_len": exp.num_clbits,
                        "q_to_c": extract_q_to_clbit_map(exp),
                    }
                )
            else:
                per_exp_meta.append({"c_reg_len": None, "q_to_c": None})

        # convert all non-schedules to schedules
        experiments = [
            (
                compiler.schedule(experiment, backend=self)
                if (type(experiment) is not pulse.ScheduleBlock)
                and (type(experiment) is not pulse.Schedule)
                else experiment
            )  # already a schedule, so don't convert
            for experiment in experiments
        ]
        # assemble schedules to PulseQobj
        with warnings.catch_warnings():
            # The method assemble is deprecated
            warnings.filterwarnings(
                "ignore", category=DeprecationWarning, module="qiskit"
            )
            qobj = assemble(
                experiments=experiments,
                backend=self,
                shots=self.options.shots,
                qubit_lo_freq=self.qubit_lo_freq,
                meas_lo_freq=self.meas_lo_freq,
                **kwargs,
            )

        # patch qobj experiments with correct circuit-level classical width + mapping
        max_slots = getattr(qobj.config, "memory_slots", 0) or 0
        for i, qexp in enumerate(qobj.experiments):
            c_reg_len = per_exp_meta[i]["c_reg_len"]
            q_to_c = per_exp_meta[i]["q_to_c"]

            if c_reg_len is None:
                c_reg_len = derive_reg_len_from_qobj_experiment(qexp)

            if c_reg_len is not None:
                # set per-experiments width, that is going to be passed back to SDK with Results/Counts obj
                if hasattr(qexp, "header") and qexp.header is not None:
                    setattr(qexp.header, "memory_slots", int(c_reg_len))

                    md = getattr(qexp.header, "metadata", None) or {}
                    md.setdefault("tergite", {})
                    md["tergite"]["c_reg_len"] = int(c_reg_len)

                    if q_to_c:
                        md["tergite"]["q_to_c"] = {
                            str(k): int(v) for k, v in q_to_c.items()
                        }
                        # also convinient inverse view
                        meas_by_clbit = [None] * int(c_reg_len)
                        for q, c in q_to_c.items():
                            if 0 <= c < int(c_reg_len):
                                meas_by_clbit[c] = q
                        md["tergite"]["meas_qubits_by-clbit"] = meas_by_clbit
                    qexp.header.metadata = md

                max_slots = max(max_slots, int(c_reg_len))

            # enforce acquire memory slots to match circuit clbits
            if q_to_c:
                rewrite_acquire_memory_slots(qexp, q_to_c)

        # keep qobj-wide memory_slots large enough for all experiments
        if hasattr(qobj, "config") and qobj.config is not None:
            setattr(qobj.config, "memory_slots", int(max_slots))

        return qobj

    def _refresh_device_properties(self):
        """Refreshes the device properties for this backend

        Internally, a fresh call to the calibrations endpoint is made to the REST API
        """
        if self.data.characterized:
            self._device_properties = api_client.get_latest_calibration(
                self.provider.account, backend_name=self.name
            )
            logging.info(f"Refreshed the device properties of '{self.name}' backend")

    def _refresh_target(self):
        """Recompiles the target for this backend

        Internally, a fresh call to the calibrations endpoint is made to the REST API
        """
        gmap = Target(num_qubits=self.data.number_of_qubits, dt=self.data.dt)
        if self.data.characterized:
            self._refresh_device_properties()

            device_compiler.add_instructions(
                backend=self,
                qubits=tuple(q for q in range(self.data.number_of_qubits)),
                coupled_qubit_idxs=self.data.coupled_qubit_idxs,
                target=gmap,
                device_properties=self.device_properties,
            )

        logging.info(f"Refreshed the target for '{self.name}' backend")
        self._target = gmap

    @property
    def meas_map(self) -> list:
        return self.data.meas_map

    @property
    def last_calibrated(self) -> Optional[str]:
        """The timestamp for when this backed was last calibrated"""
        try:
            return self.device_properties.last_calibrated
        except AttributeError:
            pass

    @functools.cached_property
    def coupling_map(self) -> CouplingMap:
        # It is required that nodes are contiguously indexed starting at 0.
        # Missed nodes will be added as isolated nodes in the coupling map.
        #
        # Also for some reason the graph has to be connected in Qiskit? Not sure why that is..
        edges = sorted(
            sorted(self.data.coupling_map, key=lambda i: i[1]), key=lambda i: i[0]
        )
        return CouplingMap(couplinglist=list(edges))

    @classmethod
    def _default_options(cls, /) -> Options:
        """This defines the default user configurable settings of this backend.

        See: help(backend.set_options)

         Returns:
            qiskit.providers.Options: A options object with
                default values set
        """
        options = Options(shots=2000)
        options.set_validator("shots", (1, TergiteBackend.max_shots))
        return options

    def _as_dict(self):
        """Converts this backend into a dictionary representation

        Returns:
            dict: a dictionary representation of this backend
        """
        obj = self.configuration().to_dict()
        obj["characterized"] = self.data.characterized
        return obj

    def __repr__(self) -> str:
        repr_list = [f"TergiteBackend object @ {hex(id(self))}:"]
        config = self._as_dict()
        for attr, value in config.items():
            repr_list.append(f"  {attr}:\t{value}".expandtabs(30))
        return "\n".join(repr_list)

    def __eq__(self, other: Any):
        if not isinstance(other, TergiteBackend):
            return False

        self_dict = self._as_dict().copy()
        other_dict = other._as_dict().copy()

        # serialize a few items that are hard to serialize
        self_dict["coupling_map"] = f"{self_dict['coupling_map']}"
        other_dict["coupling_map"] = f"{other_dict['coupling_map']}"

        self_dict["supported_instructions"] = f"{self_dict['supported_instructions']}"
        other_dict["supported_instructions"] = f"{other_dict['supported_instructions']}"

        return self_dict == other_dict


class OpenQASMBackend(TergiteBackend):
    open_pulse = False

    @functools.cached_property
    def target(self) -> Target:
        gmap = Target(num_qubits=self.data.number_of_qubits)

        # add rx, rz, delay, measure gate set
        gmap.add_instruction(circuit.library.Reset())
        gmap.add_instruction(
            circuit.library.standard_gates.RXGate(circuit.Parameter("theta"))
        )
        gmap.add_instruction(
            circuit.library.standard_gates.RZGate(circuit.Parameter("lambda"))
        )
        gmap.add_instruction(circuit.Delay(circuit.Parameter("tau")))
        gmap.add_instruction(circuit.measure.Measure())
        gmap.add_instruction(circuit.library.CZGate())

        return gmap

    def make_qobj(
        self, experiments: Union[List[QuantumCircuit], QuantumCircuit], /, **kwargs
    ) -> QasmQobj:
        """Constructs a QasmQobj from an OpenQASM circuit, or a list of them

        Raises:
            TypeError: If any of the experiments passed in not an instance
                of :class:`~qiskit.qobj.qasm_qobj.QasmQobj`

        Returns:
            qiskit.qobj.qasm_qobj.QasmQobj: A QasmQobj object transpiled from the circuits
        """
        if type(experiments) is not list:
            experiments = [experiments]
        for e in experiments:
            if not isinstance(e, QuantumCircuit):
                raise TypeError(f"Experiment {e} is not an instance of QuantumCircuit.")

        circuits = compiler.transpile(circuits=experiments, backend=self)
        with warnings.catch_warnings():
            # The class QobjExperimentHeader is deprecated
            warnings.filterwarnings(
                "ignore", category=DeprecationWarning, module="qiskit"
            )
            return assemble(experiments=circuits, shots=self.options.shots, **kwargs)

    def configuration(self) -> BackendConfiguration:
        return BackendConfiguration(
            backend_name=self.name,  # From BackendV2.
            backend_version=self.backend_version,  # From BackendV2.
            n_qubits=self.target.num_qubits,  # Number of qubits, obtained from self.target.
            basis_gates=[],  # Don't need.
            gates=[],  # Don't need.
            simulator=False,  # This is a real quantum computer.
            conditional=False,  # We cannot do conditional gate application yet.
            local=False,  # Jobs are sent over the internet.
            open_pulse=False,  # This backend just transpiles circuits.
            meas_levels=(2,),  # 0: RAW, 1: INTEGRATED, 2: DISCRIMINATED
            memory=True,  # Meas level 2 results are stored in MongoDB (?).
            max_shots=TergiteBackend.max_shots,  # From TergiteBackend.
            coupling_map=self.coupling_map,  # From TergiteBackend.
        )



from collections.abc import Iterable
from typing import Dict

from qiskit.circuit import QuantumCircuit


def as_circuit_list(experiments):
    """Always return a list of QuantumCircuit objects."""
    if isinstance(experiments, QuantumCircuit):
        return [experiments]
    if isinstance(experiments, Iterable):
        return list(experiments)
    raise TypeError("experiments must be a QuantumCircuit or iterable")


def normalise_classical_registers(
    qc: QuantumCircuit, /, *, prefer_c: bool
) -> QuantumCircuit:
    """
    Return `qc` with at most one classical register.
    By default when you initialize QuantumCircuit in Qiskit i.e. with QuantumCircuit(2, 2).
    It would create classical register named 'c'. Which you can access with measure(qbit, cbit).
    However, when you do measurements through measure_all() - it would create a different register 'meas'.
    To avoid index errors on backend where measurements are send back to SDK with meas_maps, we need to delete
    one of the registers that is not used.

    • If `prefer_c` is True and a classical register named 'c' exists and has the
      required length, measurements are rewritten to target that register
      and every unused register is dropped.

    • Otherwise (or if 'c' is unusable), any register that contains no
      measured bits is removed; the remaining register(s) stay unchanged.

    The function never deletes a register that holds at least one bit
    referenced in an instruction, so partial-measurement circuits are
    preserved as-is.
    """
    # classifying the existing registers
    # used bits are the ones we can find in the instructions
    used_bits = {cb for inst in qc.data for cb in inst.clbits}

    # it is assumed that a circuit does at least one measurement
    # we can then filter a creg that has at least one bit used
    reg_used = {creg: any(cb in used_bits for cb in creg) for creg in qc.cregs}

    # early exit – nothing to rewrite or delete
    if sum(reg_used.values()) == 1 and all(reg_used.values()):
        return qc

    # decide which register we want to keep
    keep_creg = None
    if prefer_c:
        for creg in qc.cregs:
            if creg.name == "c" and len(creg) >= len(used_bits):
                keep_creg = creg
                break
    if keep_creg is None:
        # keep the first register that actually contains a measurement
        keep_creg = next(creg for creg, used in reg_used.items() if used)

    # build a cleaned‑up circuit
    qc_new = QuantumCircuit(qc.num_qubits)
    qc_new.add_register(keep_creg)

    # mapping old clbit → new clbit (only for bits that are used)
    bit_map = {
        cb: keep_creg[i]
        for i, cb in enumerate(
            sorted(used_bits, key=lambda x: (x._register.name, x._index))
        )
    }

    # copy/rewrite every instruction
    for inst in qc.data:
        new_cargs = [bit_map[cb] for cb in inst.clbits]
        qc_new.append(inst.operation, inst.qubits, new_cargs)

    qc_new.global_phase = qc.global_phase
    qc_new.metadata = getattr(qc, "metadata", None)
    qc_new.name = getattr(qc, "name", None)
    return qc_new


def extract_q_to_clbit_map(circ: QuantumCircuit) -> Dict[int, int]:
    """Map circuit qubit-index -> flattened classical-bit index."""
    q_to_c: Dict[int, int] = {}
    for inst, qargs, cargs in circ.data:
        if inst.name != "measure":
            continue
        # measure is 1 qubit -> 1 clbit
        q_i = circ.find_bit(qargs[0]).index
        c_i = circ.find_bit(cargs[0]).index
        q_to_c[q_i] = c_i
    return q_to_c



# This code is part of Tergite
#
# (C) Chalmers Next Labs (2024)
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# ---------------- ALTERATION NOTICE ---------------- #
# This code was modified from the original by:
# Chalmers Next Labs 2024
#
# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#

"""
=========================
Qobj (:mod:`qiskit.qobj`)
=========================

.. currentmodule:: qiskit.qobj

Base
====

.. autosummary::
   :toctree: ../stubs/

   QobjExperimentHeader
   QobjHeader

Qasm
====

.. autosummary::
   :toctree: ../stubs/

   QasmQobj
   QasmQobjInstruction
   QasmQobjExperimentConfig
   QasmQobjExperiment
   QasmQobjConfig
   QasmExperimentCalibrations
   GateCalibration

Pulse
=====

.. autosummary::
   :toctree: ../stubs/

   PulseQobj
   PulseQobjInstruction
   PulseQobjExperimentConfig
   PulseQobjExperiment
   PulseQobjConfig
   QobjMeasurementOption
   PulseLibraryItem
"""

from .common import QobjExperimentHeader, QobjHeader
from .pulse_qobj import (
    PulseLibraryItem,
    PulseQobj,
    PulseQobjConfig,
    PulseQobjExperiment,
    PulseQobjExperimentConfig,
    PulseQobjInstruction,
    QobjMeasurementOption,
)
from .qasm_qobj import (
    GateCalibration,
    QasmExperimentCalibrations,
    QasmQobj,
    QasmQobjConfig,
    QasmQobjExperiment,
    QasmQobjExperimentConfig,
    QasmQobjInstruction,
)



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Alteration Notice
# -----------------
# This code was refactored from the original by:
#
# Martin Ahindura, 2023
"""Defines the ProviderFactory class that contains multiple Tergite Providers."""
import pathlib
from collections import OrderedDict
from typing import List, Optional

from ..services.configs import TERGITERC_FILE, AccountInfo, Tergiterc
from .provider import Provider


class ProviderFactory:
    """Container of multiple Tergite Provider's, retrievable by service name

    It can be passed a custom path to an rc file: default is `$HOME/.qiskit/tergiterc`
    """

    def __init__(self, rc_file: pathlib.Path = TERGITERC_FILE):
        self._providers = OrderedDict()
        self._tergiterc = Tergiterc(rc_file=rc_file)

        # get a list of provider account objects from tergiterc
        accounts_list = self._tergiterc.load_accounts()
        if not accounts_list:
            print(
                "Warning: No stored provider account found in $HOME/.qiskit/tergiterc"
            )

        # for each item in the accounts list create a Provider object
        for account in accounts_list:
            self._providers[account.service_name] = Provider(account)

    def use_provider_account(
        self,
        service_name: str,
        url: str,
        token: Optional[str] = None,
        save: bool = False,
    ) -> Provider:
        """Initializes a new Provider basing on the account passed and returns it.

        Args:
            service_name: the name of the service
            url: the API URL for the given service
            token: the API token to be used to connect to the service
            save: whether the account should be persisted to the tergiterc file

        Returns:
            A new instance of the
                :class:`~tergite.providers.tergite.provider.Provider`
                for the given account
        """
        account = AccountInfo(service_name=service_name, url=url, token=token)
        if save:
            self._tergiterc.save_accounts([account])

        new_provider = Provider(account)
        self._providers[account.service_name] = new_provider

        return new_provider

    def providers(self) -> List[str]:
        """Retrieves the service names of providers in this collection

        Returns:
            list of service names of providers in this collection
        """
        return list(self._providers.keys())

    def get_provider(self, service_name=None) -> Optional[Provider]:
        """Retrieves a given provider by service name

        Args:
            service_name: the name by which the given provider is saved

        Returns:
            the provider of the given service name
                if it exists or None if it doesn't
        """
        providers = self._providers

        if not providers:
            raise Exception(
                "No provider account is available. Provide one via Tergite.use_provider_account(..)."
            )

        # get by name or default to the first entry
        return providers.get(service_name, next(iter(providers.values())))



# This code is part of Tergite
#
# (C) Copyright Chalmers Next Labs 2025
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Module containing the types for data transfer"""
from __future__ import annotations

import enum
import json
from typing import Any, Dict, List, Literal, Optional, Tuple, TypeAlias, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    SerializationInfo,
    field_serializer,
    field_validator,
)
from pydantic.main import IncEx

from tergite.compat.qiskit.qobj import PulseQobj
from tergite.compat.qiskit.qobj.encoder import IQXJsonEncoder

IQPoint: TypeAlias = Tuple[float, float]  # [re, im]  (len = 2)
IQMemory: TypeAlias = List[List[List[IQPoint]]]  # exp → shot → IQ points
HexMemory: TypeAlias = List[
    List[str]
]  # experiments x reps/shots x str(channels x bits)


class TergiteBackendConfig(BaseModel):
    """backend configuration got from the remote server"""

    model_config = ConfigDict(extra="allow")

    # Fields without default values
    name: str
    version: str
    number_of_qubits: int
    is_online: bool
    basis_gates: List[str]
    coupling_map: List[Tuple[int, int]]
    coordinates: List[Tuple[int, int]]
    is_simulator: bool
    characterized: bool
    open_pulse: bool
    meas_map: List[List[int]]

    # Fields with default values
    id: Optional[str] = None
    last_online: Optional[str] = None
    description: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    number_of_couplers: int = 0
    number_of_resonators: int = 0
    dt: Optional[float] = None
    dtm: Optional[float] = None
    coupling_dict: Dict[str, Tuple[str, str]] = {}
    coupled_qubit_idxs: Tuple[Tuple[int, int], ...] = []
    qubit_ids_coupler_dict: Dict[Tuple[int, int], int] = {}
    qubit_ids_coupler_map: List[Tuple[Tuple[int, int], int]] = []
    qubit_ids: List[str] = []
    meas_lo_freq: Optional[List[int]] = None
    qubit_lo_freq: Optional[List[int]] = None
    gates: Optional[Dict[str, Any]] = None
    is_active: Optional[bool] = None

    def model_post_init(self, _context: Any):
        """Run after initialization of the model"""
        # qubits_coupler_map is a list of key,value tuples for the qubit_ids_coupler_dict
        # It is so because a dict with tuples as keys is not JSON serializable
        self.qubit_ids_coupler_dict = {
            tuple(k): v for (k, v) in self.qubit_ids_coupler_map
        }

        # the coupling_map sometimes has qubits connected to themselves e.g. [0, 0] when there are
        # no couplers so we need to filter these out to get the qubits that are actually coupled
        self.coupled_qubit_idxs = tuple(
            [pair for pair in self.coupling_map if pair[0] != pair[1]]
        )


class CalibrationValue(BaseModel):
    """A calibration value"""

    model_config = ConfigDict(extra="allow")

    value: Union[float, str, int]
    date: Optional[str] = None
    unit: str = ""


class QubitCalibration(BaseModel):
    """Schema for the calibration data of the qubit"""

    model_config = ConfigDict(extra="allow")

    t1_decoherence: Optional[CalibrationValue] = None
    t2_decoherence: Optional[CalibrationValue] = None
    frequency: Optional[CalibrationValue] = None
    anharmonicity: Optional[CalibrationValue] = None
    readout_assignment_error: Optional[CalibrationValue] = None
    # parameters for x gate
    pi_pulse_amplitude: Optional[CalibrationValue] = None
    pi_pulse_duration: Optional[CalibrationValue] = None
    pi_pulse_motzoi: Optional[CalibrationValue] = None
    pulse_type: Optional[CalibrationValue] = None
    pulse_sigma: Optional[CalibrationValue] = None
    id: Optional[int] = None
    index: Optional[CalibrationValue] = None
    x_position: Optional[CalibrationValue] = None
    y_position: Optional[CalibrationValue] = None
    xy_drive_line: Optional[CalibrationValue] = None
    z_drive_line: Optional[CalibrationValue] = None


class ResonatorCalibration(BaseModel):
    """Schema for the calibration data of the resonator"""

    model_config = ConfigDict(extra="allow")

    acq_delay: Optional[CalibrationValue] = None
    acq_integration_time: Optional[CalibrationValue] = None
    frequency: Optional[CalibrationValue] = None
    pulse_amplitude: Optional[CalibrationValue] = None
    pulse_delay: Optional[CalibrationValue] = None
    pulse_duration: Optional[CalibrationValue] = None
    pulse_type: Optional[CalibrationValue] = None
    id: Optional[int] = None
    index: Optional[CalibrationValue] = None
    x_position: Optional[CalibrationValue] = None
    y_position: Optional[CalibrationValue] = None
    readout_line: Optional[CalibrationValue] = None


class CouplersCalibration(BaseModel):
    """Schema for the calibration data of the coupler"""

    model_config = ConfigDict(extra="allow")

    frequency: Optional[CalibrationValue] = None
    frequency_detuning: Optional[CalibrationValue] = None
    anharmonicity: Optional[CalibrationValue] = None
    coupling_strength_02: Optional[CalibrationValue] = None
    coupling_strength_12: Optional[CalibrationValue] = None
    cz_pulse_amplitude: Optional[CalibrationValue] = None
    cz_pulse_dc_bias: Optional[CalibrationValue] = None
    cz_pulse_phase_offset: Optional[CalibrationValue] = None
    cz_pulse_duration_before: Optional[CalibrationValue] = None
    cz_pulse_duration_rise: Optional[CalibrationValue] = None
    cz_pulse_duration_constant: Optional[CalibrationValue] = None
    control_rz_lambda: Optional[CalibrationValue] = None
    target_rz_lambda: Optional[CalibrationValue] = None
    pulse_type: Optional[CalibrationValue] = None
    id: Optional[int] = None


class DeviceCalibration(BaseModel):
    """Schema for the calibration data of a given device"""

    model_config = ConfigDict(extra="allow")

    name: str
    version: str
    qubits: List[QubitCalibration]
    resonators: Optional[List[ResonatorCalibration]] = None
    couplers: Optional[List[CouplersCalibration]] = None
    discriminators: Optional[Dict[str, Any]] = None
    last_calibrated: str


class CreatedJobResponse(BaseModel):
    """The response when a new job is created"""

    model_config = ConfigDict(
        extra="allow",
    )

    job_id: str
    upload_url: str
    access_token: str


class RemoteJobResult(BaseModel):
    """The results of the job"""

    model_config = ConfigDict(
        extra="allow",
    )

    memory: Union[HexMemory, IQMemory] = []


class RemoteJobStatus(str, enum.Enum):
    PENDING = "pending"
    SUCCESSFUL = "successful"
    FAILED = "failed"
    EXECUTING = "executing"
    CANCELLED = "cancelled"


class RemoteJob(BaseModel):
    """the job schema"""

    model_config = ConfigDict(
        extra="allow",
    )

    device: str
    calibration_date: str
    job_id: str
    project_id: Optional[str] = None
    user_id: Optional[str] = None
    status: RemoteJobStatus
    failure_reason: Optional[str] = None
    cancellation_reason: Optional[str] = None
    download_url: Optional[str] = None
    result: Optional[RemoteJobResult] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None


class JobFileParams(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    qobj: PulseQobj

    @field_serializer("qobj")
    def serialize_qobj(self, qobj: PulseQobj, _info: SerializationInfo):
        """Converts qobj into a dict"""
        qobj_dict = qobj.to_dict()

        if _info.mode_is_json():
            return json.dumps(qobj_dict, cls=IQXJsonEncoder)
        return qobj_dict

    @field_validator("qobj", mode="before")
    @classmethod
    def parse_qobj(cls, v):
        """Parses the qobject from dict/str to Qobj"""
        if isinstance(v, PulseQobj):
            return v
        elif isinstance(v, dict):
            return PulseQobj.from_dict(v)
        elif isinstance(v, str):
            return PulseQobj.from_dict(json.loads(v))

        raise TypeError(f"Invalid type for PulseQobj: {type(v)}")


class JobFile(BaseModel):
    """The expected structure of the job file"""

    model_config = ConfigDict(extra="forbid")

    job_id: str
    params: JobFileParams



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Alteration Notice
# -----------------
# This code was refactored from the original by:
#
# Martin Ahindura, 2023
# Chalmers Next Labs, 2025

"""Entry into the configuration service"""
import dataclasses
import pathlib
import re
from configparser import ConfigParser
from typing import List, Optional

from .dtos import AccountInfo

TERGITERC_FILE = pathlib.Path.home() / ".qiskit" / "tergiterc"


class Tergiterc:
    """the Configuration parser for tergiterc files"""

    def __init__(self, rc_file: pathlib.Path = TERGITERC_FILE):
        """Initializes a Tergiterc instance

        The instance initialized, saves to and retrieves its
        data from file_path

        Args:
            rc_file: the path where the tergiterc file is saved, defaults to ``$HOME/.qiskit/tergiterc``
        """
        self._file_path = rc_file
        self._parser = Tergiterc._get_parser(rc_file)

    def load_accounts(self) -> List["AccountInfo"]:
        """Retrieves the accounts from the tergiterc file

        Returns:
            list of instances of
                :class:`~tergite.providers.tergite.account.AccountInfo`
                as read from the tergiterc file
        """

        account_fields = {field.name: True for field in dataclasses.fields(AccountInfo)}
        accounts = []

        parser = self._parser
        if not parser:
            return accounts

        sections_all = parser.sections()
        for section in sections_all:
            if section.startswith("service"):
                if not parser.has_option(section, "url"):
                    print(
                        f"Warning: Skipping account provider '{section}'. Invalid configuration."
                    )
                    continue

                section_items = dict(parser.items(section))
                service_name = section.split(" ", 1)[1].strip()

                new_account = AccountInfo(service_name, **section_items)
                accounts.append(new_account)

        return accounts

    def save_accounts(self, accounts: List["AccountInfo"]):
        """Saves the accounts into the tergiterc file

        Args:
            accounts: the list of instances of
                :class:`~tergite.providers.tergite.account.AccountInfo`
                to save

        Raises:
            Exception: If no accounts are passed
        """
        if not accounts:
            raise Exception("Cannot save account(s). None given.")

        if not self._parser:
            self._parser = ConfigParser()

        for account in accounts:
            # section
            section_name = "service " + account.service_name
            self._parser.add_section(section_name)

            # account configuration
            config = account.to_dict()
            config.pop("service_name")  # remove 'service_name'

            for key, value in config.items():
                self._parser.set(section_name, str(key), str(value))

        with self._file_path.open("w") as dest:
            self._parser.write(dest)

    @staticmethod
    def _get_parser(rc_file: pathlib.Path) -> Optional[ConfigParser]:
        """Initializes a :class:`configparser.ConfigParser` instance that
        has read from rc_file

        It returns None if rc_file does not exist.

        Returns:
            Optional[configparser.ConfigParser]: the ConfigParser with the
                rc_file loaded in it if the file exists
        """
        if not rc_file.exists():
            return None

        parser = ConfigParser()
        parser.SECTCRE = re.compile(r"\[ *(?P<header>[^]]+?) *\]")
        parser.read(rc_file)

        return parser



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Alteration Notice
# -----------------
# This code was modified from the original by:
#
# - Martin Ahindura 2023
# - Stefan Hill 2023
"""Handles the creation of schedules for the devices of type TergiteBackend"""
from typing import TYPE_CHECKING, Iterable

import numpy as np
import qiskit.circuit as circuit
import qiskit.pulse as pulse
from qiskit.pulse.library import SymbolicPulse
from sympy import symbols

from .functions import delta_t_function_sympy

if TYPE_CHECKING:
    from ...types.backend import TergiteBackend
    from ..api_client.dtos import DeviceCalibration


def rx(
    backend: "TergiteBackend",
    qubits: Iterable,
    rx_theta: circuit.Parameter,
    device_properties: "DeviceCalibration",
) -> pulse.ScheduleBlock:
    """Creates a rotation around the x-axis on the Bloch sphere for the list of qubits in the backend.

    It returns a backend-specific schedule which implements a
    rotation on a set of qubits around the x-axis on the Bloch sphere.
    The schedule is parameterised by θ [rad], which specifies the angle.

    Args:
        backend: the backend for which the schedule is to be created.
        qubits: the set of qubits to be affected.
        rx_theta: the theta parameter for the circuit
        device_properties: the device parameters of the backend

    Returns:
        qiskit.pulse.ScheduleBlock: the schedule implementing the rotation
    """
    qubit = device_properties.qubits

    sched = pulse.ScheduleBlock(name=f"RX(θ, {qubits})")
    for q in qubits:
        sched += pulse.SetFrequency(
            qubit[q].frequency.value,
            channel=backend.drive_channel(q),
        )
        duration = round(qubit[q].pi_pulse_duration.value / backend.dt)
        calibration_sigma = qubit[q].pulse_sigma.value
        sigma = (
            round(calibration_sigma / backend.dt)
            if calibration_sigma
            else round(duration / 4)
        )
        amplitude = rx_theta / np.pi * qubit[q].pi_pulse_amplitude.value
        motzoi_params = device_properties.qubits[q].pi_pulse_motzoi

        motzoi = motzoi_params.value if motzoi_params else 0.0
        sched += pulse.Play(
            pulse.Drag(
                duration=duration,
                amp=amplitude,
                sigma=sigma,
                beta=motzoi,
                name=f"RX q{q}",
            ),
            channel=backend.drive_channel(q),
        )

    return sched


def rz(
    backend: "TergiteBackend", qubits: Iterable, rz_lambda: circuit.Parameter
) -> pulse.ScheduleBlock:
    """Creates a rotation around the z-axis on the Bloch sphere for the list of qubits in the backend.

    It returns a backend-specific schedule which implements a rotation on a set
    of qubits around the z-axis on the Block sphere. The schedule is
    parameterised by λ [rad], which specifies the angle.

    Args:
        backend: the backend for which the schedule is to be created.
        qubits: the set of qubits to be affected.
        rz_lambda: the lambda parameter for the circuit

    Returns:
        qiskit.pulse.ScheduleBlock: the schedule implementing the rotation
    """
    sched = pulse.ScheduleBlock(name=f"RZ(λ, {qubits})")
    for q in qubits:
        sched += pulse.ShiftPhase(
            rz_lambda, channel=backend.drive_channel(q), name=f"RZ q{q}"
        )
    return sched


def cz(
    backend: "TergiteBackend",
    control_qubit_idxs: Iterable,
    target_qubit_idxs: Iterable,
    device_properties: "DeviceCalibration",
) -> pulse.ScheduleBlock:
    """Two qubit CNOT gate building block. TODO: Add doc comment."""
    # FIXME: Why choose to pass control_qubits and target_qubits as Iterables yet
    #    when being called, only as single value is passed
    sched = pulse.ScheduleBlock(
        name=f"CZ(θ, control={control_qubit_idxs}, target={target_qubit_idxs})"
    )

    for control_qubit_idx, target_qubit_idx in zip(
        control_qubit_idxs, target_qubit_idxs
    ):
        # FIXME: there is a an assumption how we choose target and control qubits for local phases
        # This should be replaced with proper explicit labels for target and control qubit ids per coupler
        # The assumption: when we have two qubits q_x and q_y coupled by a coupler
        # the control qubit id from calibration config is even, target id is odd
        control_qubit_id_from_calibration = device_properties.qubits[
            control_qubit_idx
        ].id
        is_control_qubit_even = control_qubit_id_from_calibration % 2 == 0
        cz_control_qubit_idx = (
            control_qubit_idx if is_control_qubit_even else target_qubit_idx
        )
        cz_target_qubit_idx = (
            target_qubit_idx if is_control_qubit_even else control_qubit_idx
        )

        control_qubit = device_properties.qubits[cz_control_qubit_idx]
        target_qubit = device_properties.qubits[cz_target_qubit_idx]

        # Get control channels between control and target qubits

        control_channels = backend.control_channel((control_qubit.id, target_qubit.id))

        c_props = [
            x for x in device_properties.couplers if x.id == control_channels[0].index
        ]
        c_props = c_props[0] if c_props else None

        # TODO: raise error if c_props is none

        f1 = min(_val(target_qubit.frequency), _val(control_qubit.frequency))
        f0 = max(_val(target_qubit.frequency), _val(control_qubit.frequency))

        alpha0 = _val(c_props.anharmonicity, default=-0.17e9)

        f2 = _val(c_props.frequency)  # Coupler frequency
        detuning = _val(
            c_props.frequency_detuning, default=0
        )  # detuning in radial frequency

        args = {
            "delta_0": _val(c_props.cz_pulse_amplitude),  # Maximum delta value
            "Theta": _val(c_props.cz_pulse_dc_bias, default=0.0),  # DC bias term
            "omega_c0": 2 * np.pi * f2,  # Maximum frequency in Hz
            "omega_Phi": detuning
            + 2 * np.pi * (f1 - f0 - alpha0),  # Transition frequency in Hz
            "phi": _val(c_props.cz_pulse_phase_offset, default=0.0),  # Phase offset
            "t_w": _val(
                c_props.cz_pulse_duration_before, default=4e-9
            ),  # s, duration before pulse
            "t_rf": _val(c_props.cz_pulse_duration_rise, default=4e-9),  # s, rise time
            "t_p": _val(
                c_props.cz_pulse_duration_constant
            ),  # s, constant pulse duration
        }

        # we now set wait time before and after pulse explicitly for both sim and real backends
        t_gate = args["t_p"] + args["t_rf"]
        # required param for pulse, for display purposes delta_0 is equivalent to max_amplitude
        amp = args["delta_0"]
        # this is for display purposes, as we overriding frequency modulation in backend
        freq = int(f2)

        args["amp"] = amp
        args["freq"] = freq
        control_rz_lambda = c_props.control_rz_lambda.value
        target_rz_lambda = c_props.target_rz_lambda.value
        cz_gate_duration = round(t_gate / backend.dt)
        cz_wait_duration = round(args["t_w"] / backend.dt)

        cz_gate = wacqt_cz_gate(
            duration=cz_gate_duration, name="cz_pulse", numerical_args=args
        )
        sched += pulse.SetFrequency(
            c_props.frequency.value,
            channel=control_channels[0],
        )
        sched += pulse.Delay(channel=control_channels[0], duration=cz_wait_duration)
        sched += pulse.Play(cz_gate, control_channels[0])
        sched += pulse.Delay(channel=control_channels[0], duration=cz_wait_duration)

        if control_rz_lambda and target_rz_lambda:
            sched += pulse.ShiftPhase(
                control_rz_lambda,
                channel=backend.drive_channel(cz_control_qubit_idx),
                name=f"RZ q{cz_control_qubit_idx}",
            )
            sched += pulse.ShiftPhase(
                target_rz_lambda,
                channel=backend.drive_channel(cz_target_qubit_idx),
                name=f"RZ q{cz_target_qubit_idx}",
            )

    return sched


def measure(
    backend: "TergiteBackend",
    qubits: Iterable,
    device_properties: "DeviceCalibration",
) -> pulse.ScheduleBlock:
    """Creates a measurement on the list of qubits in the given backend.

    It returns a backend-specific schedule which implements
    a measurement on a set of qubits.

    Args:
        backend: the backend for which the schedule is to be created.
        qubits: the set of qubits to be affected.
        device_properties: the device parameters of the backend

    Returns:
        qiskit.pulse.ScheduleBlock: the schedule implementing the measurement
    """
    readout_resonator_props = device_properties.resonators

    sched = pulse.ScheduleBlock(name=f"Measure({qubits})")
    for q in qubits:
        readout_resonator = readout_resonator_props[q]
        sched += pulse.SetFrequency(
            readout_resonator.frequency.value, channel=backend.measure_channel(q)
        )
        sched += pulse.Delay(
            channel=backend.measure_channel(q), duration=round(4e-9 / backend.dt)
        )
        sched += pulse.Delay(
            channel=backend.acquire_channel(q), duration=round(4e-9 / backend.dt)
        )
        sched += pulse.Play(
            pulse.Constant(
                amp=readout_resonator.pulse_amplitude.value,
                duration=round(readout_resonator.pulse_duration.value / backend.dt),
                name=f"Readout q{q}",
            ),
            channel=backend.measure_channel(q),
        )
        sched += pulse.Delay(
            duration=round(readout_resonator.acq_delay.value / backend.dt),
            channel=backend.acquire_channel(q),
            name=f"Time of flight q{q}",
        )
        sched += pulse.Acquire(
            duration=round(readout_resonator.acq_integration_time.value / backend.dt),
            channel=backend.acquire_channel(q),
            mem_slot=backend.memory_slot(q),
            name=f"Integration window q{q}",
        )
    return sched


def delay(
    backend: "TergiteBackend",
    qubits: Iterable,
    delay_tau: circuit.Parameter,
    delay_str: str = "Delay",
) -> pulse.ScheduleBlock:
    """Creates a delay on the set of qubits in the backend.

    It returns a backend-specific schedule which implements a delay
    operation on a set of qubits.
    The schedule is parameterised by τ [ns], which specifies the delay duration.

    Args:
        backend: the backend for which the schedule is to be created.
        qubits: the set of qubits to be affected.
        delay_tau: the tau parameter for the circuit
        delay_str: the delay string for the instruction set

    Returns:
        qiskit.pulse.ScheduleBlock: the schedule implementing the delay
    """
    sched = pulse.ScheduleBlock(name=f"{delay_str}({qubits}, τ)")
    for q in qubits:
        sched += pulse.Delay(
            duration=delay_tau,
            channel=backend.drive_channel(q),
            name=f"{delay_str} q{q}",
        )
        sched += pulse.Delay(
            duration=delay_tau,
            channel=backend.measure_channel(q),
            name=f"{delay_str} q{q}",
        )
        sched += pulse.Delay(
            duration=delay_tau,
            channel=backend.acquire_channel(q),
            name=f"{delay_str} q{q}",
        )
    return sched


def wacqt_cz_gate(duration, name, numerical_args):
    # define the time variable
    t = symbols("t", real=True)

    # define symbolic variables (can also be passed as parameters)
    symbolic_args = {
        "t_w": symbols("t_w", real=True),
        "t_rf": symbols("t_rf", real=True),
        "t_p": symbols("t_p", real=True),
        "delta_0": symbols("delta_0", real=True),
    }

    # create the symbolic expression
    envelope_expr = delta_t_function_sympy(t, symbolic_args)

    # substitute numerical values into the symbolic expression
    if numerical_args:
        envelope_expr = envelope_expr.subs(numerical_args)

    numerical_args["amp"] = [numerical_args["amp"], 0]
    # create the SymbolicPulse instance
    instance = SymbolicPulse(
        pulse_type="Wacqt_cz_gate_pulse",
        duration=duration,
        parameters=numerical_args,
        envelope=envelope_expr,
        name=name,
    )

    return instance


def _val(x, default=None):
    """Unwrap calibration parameters safely."""
    if x is None:
        return default
    v = getattr(x, "value", x)
    return default if v is None else v



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
# (C) Copyright Axel Andersson 2022
# (C) Copyright Martin Ahindura 2023
# (C) Copyright Chalmers Next Labs 2025
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Alteration Notice
# -----------------
# This code was refactored from the original by:
#
# Martin Ahindura, 2023, 2025
# Adilet Tuleuov, 2024

"""Defines the Qiskit provider with which to access the Tergite Quantum Computers"""
import functools
from typing import Any, Callable, Dict, List, Union

from qiskit.providers.exceptions import QiskitBackendNotFoundError
from qiskit.providers.providerutils import filter_backends

from ..services import api_client
from ..services.configs import AccountInfo
from ..utils.job_file import extract_job_metadata, extract_job_qobj
from .backend import OpenPulseBackend, OpenQASMBackend
from .job import STATUS_MAP, Job


class Provider:
    """The Qiskit Provider with which to access the Tergite quantum computers"""

    def __init__(self, /, account: "AccountInfo"):
        """Initializes the Provider

        Args:
            account: the instance of the
                :class:tergite.providers.tergite.account.AccountInfo`
                with which to connect to the Tergite API
        """
        super().__init__()
        self.account = account
        self._malformed_backends = {}

    def backends(
        self, /, name: str = None, filters: Callable[[Any], bool] = None, **kwargs
    ) -> List[Union[OpenPulseBackend, OpenQASMBackend]]:
        """Filters the available backends of this provider.

        Args:
            name: the name of the backend
            filters: a callable to filter the backends with
            kwargs: kwargs to match the available backends with

        Returns:
            A list of instantiated and available OpenPulseBackend, or OpenPulseBackend backends,
                that match the given filter
        """
        available_backends = self.available_backends.values()
        if name:
            kwargs["backend_name"] = name

        return filter_backends(available_backends, filters=filters, **kwargs)

    @functools.cached_property
    def available_backends(
        self, /
    ) -> Dict[str, Union[OpenPulseBackend, OpenQASMBackend]]:
        """Dictionary of all available backends got from the API"""
        backends = dict()
        backend_configs = api_client.get_backend_configs(self.account)

        for backend_conf in backend_configs:
            if backend_conf.open_pulse:
                obj = OpenPulseBackend(
                    data=backend_conf, provider=self, base_url=self.account.url
                )
            else:
                obj = OpenQASMBackend(
                    data=backend_conf, provider=self, base_url=self.account.url
                )
            backends[obj.name] = obj

        return backends

    def job(self, job_id: str) -> Job:
        """Retrieve a runtime job given a job id

        Args:
            job_id: the ID of the job

        Returns:
            an instance of `Job` corresponding to the given job id

        Raises:
            RuntimeError: Job: {job_id} has no download_url
        """
        account = self.account
        remote_data = api_client.get_remote_job_data(account, job_id=job_id)
        if remote_data.download_url is None:
            raise RuntimeError(f"Job: {job_id} has no download_url")

        download_url = remote_data.download_url
        access_token = remote_data.access_token
        backend = self.get_backend(name=remote_data.device)
        logfile = api_client.download_job_logfile(
            job_id=job_id, url=download_url, access_token=access_token
        )
        raw_status = remote_data.status
        status = STATUS_MAP.get(raw_status)
        calibration_date = remote_data.calibration_date

        metadata = extract_job_metadata(logfile)
        payload = extract_job_qobj(logfile)

        job = Job(
            backend=backend,
            job_id=job_id,
            payload=payload,
            upload_url=None,
            download_url=download_url,
            logfile=logfile,
            status=status,
            remote_data=remote_data,
            calibration_date=calibration_date,
            access_token=access_token,
            **metadata,
        )
        self._sync_backend_options_from_job(job)
        return job

    def __str__(self, /):
        return "Tergite: Provider"

    def __repr__(self, /):
        return "<{} from Tergite Qiskit>".format(self.__class__.__name__)

    def _sync_backend_options_from_job(self, job):
        # Prefer payload (source of truth), fallback to metadata
        shots = None
        if getattr(job, "payload", None) is not None:
            shots = getattr(getattr(job.payload, "config", None), "shots", None)
        if shots is None:
            shots = job.metadata.get("shots")

        if shots is not None:
            job.backend().set_options(shots=int(shots))

    # The below code is part of Qiskit.
    #
    # (C) Copyright IBM 2020.
    #
    # This code is licensed under the Apache License, Version 2.0. You may
    # obtain a copy of this license in the LICENSE.txt file in the root directory
    # of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
    #
    # Any modifications or derivative works of this code must retain this
    # copyright notice, and modified files need to carry a notice indicating
    # that they have been altered from the originals.
    #
    # This code is part of Tergite
    #
    # (C) Chalmers Next Labs (2024)
    #
    # This code is licensed under the Apache License, Version 2.0. You may
    # obtain a copy of this license in the LICENSE.txt file in the root directory
    # of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
    #
    # Any modifications or derivative works of this code must retain this
    # copyright notice, and modified files need to carry a notice indicating
    # that they have been altered from the originals.
    #
    # ---------------- ALTERATION NOTICE ---------------- #
    # This code has been derived from the Qiskit abstract ProviderV1 class

    def get_backend(
        self, name=None, **kwargs
    ) -> Union[OpenPulseBackend, OpenQASMBackend]:
        """Return a single backend matching the specified filtering.

        Args:
            name (str): name of the backend.
            **kwargs: dict used for filtering.

        Returns:
            Backend: a backend matching the filtering.

        Raises:
            QiskitBackendNotFoundError: if no backend could be found or
                more than one backend matches the filtering criteria.
        """
        # As Qiskit ProviderV1 class was deprecated
        # See: https://github.com/Qiskit/qiskit/pull/12145

        # This class was migrated from Abstract ProviderV1 class
        # From: https://github.com/Qiskit/qiskit/blob/1.2.2/qiskit/providers/provider.py#L53

        backends = self.backends(name, **kwargs)
        if len(backends) > 1:
            raise QiskitBackendNotFoundError(
                "More than one backend matches the criteria"
            )
        if not backends:
            raise QiskitBackendNotFoundError("No backend matches the criteria")

        return backends[0]



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
# (C) Copyright Axel Andersson 2022
# (C) Copyright Martin Ahindura 2023
# (C) Copyright Chalmers Next Labs 2025
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Alteration Notice
# -----------------
# This code was refactored from the original by:
#
# Martin Ahindura, 2023
# Adilet Tuleuov, 2024
"""Client for accessing the API"""
import json
import shutil
import tempfile
from json import JSONDecodeError
from pathlib import Path
from textwrap import indent
from typing import TYPE_CHECKING, Any, Dict, List, Optional
from uuid import uuid4

import requests
from pydantic import ValidationError
from requests import Response

from ...compat.qiskit.qobj.encoder import IQXJsonEncoder
from .dtos import (
    CreatedJobResponse,
    DeviceCalibration,
    JobFile,
    RemoteJob,
    TergiteBackendConfig,
)

if TYPE_CHECKING:
    from ..configs import AccountInfo


def get_backend_configs(account: "AccountInfo") -> List[TergiteBackendConfig]:
    """Retrieves the backend configs from the remote API

    These configs are used to construct Backend objects
    It ignores any malformed configs.
    The expectation is no malformed config is sent over to the client.

    Args:
        account: the provider account to use

    Raises:
        RuntimeError: Error retrieving backends: {detail}
        ValidationError: if records returned does not conform to TergiteBackendConfig schema

    Returns:
        list of TergiteBackendConfig got from the remote API
    """
    url = f"{account.url}/devices/"
    headers = _get_auth_headers(account)

    response = requests.get(url, headers=headers)
    if not response.ok:
        error_msg = _get_err_text(response)
        raise RuntimeError(f"Error retrieving backends: {error_msg}")

    records = response.json()["data"]
    parsed_records = []
    for item in records:
        try:
            parsed_records.append(TergiteBackendConfig.model_validate(item))
        except ValidationError:
            pass

    return parsed_records


def get_latest_calibration(
    account: "AccountInfo", backend_name: Optional[str] = None
) -> DeviceCalibration:
    """Retrieves the latest backend calibration data and returns Calibration objects

    Args:
        account: the provider account to use to access the API
        backend_name: the name of the backend

    Returns:
        the DeviceCalibration for the given backend

    Raises:
        RuntimeError: Failed to get device calibrations for '{backend_name}': {error_msg}
        ValidationError: if records returned does not conform to DeviceCalibration schema
        ValueError: Error parsing device calibration data for '{backend_name}': {detail}
    """
    url = f"{account.url}/calibrations/{backend_name}"
    headers = _get_auth_headers(account)
    response = requests.get(url, headers=headers)

    if not response.ok:
        error_msg = _get_err_text(response)
        raise RuntimeError(
            f"failed to get device calibrations for '{backend_name}': {error_msg}"
        )

    raw_data = response.json()
    return DeviceCalibration.model_validate(raw_data)


def register_job(
    account: "AccountInfo",
    backend_name: str,
    calibration_date: Optional[str] = None,
    **kwargs,
) -> CreatedJobResponse:
    """Registers the job on the remote API

    Args:
        account: the provider account needed to access the API
        backend_name: the name of the backend
        calibration_date: the current last_calibrated timestamp for the given backend

    Returns:
        a dict with the response from the API

    Raises:
        RuntimeError: unable to register job at the remote API: {detail}
    """
    url = f"{account.url}/jobs/"
    headers = _get_auth_headers(account)
    payload = {"device": backend_name, "calibration_date": calibration_date}
    response = requests.post(url, headers=headers, json=payload)
    if not response.ok:
        err_msg = _get_err_text(response)
        raise RuntimeError(f"unable to register job at the remote API: {err_msg}")

    raw_data = response.json()
    return CreatedJobResponse.model_validate(raw_data)


def send_job_file(
    account: "AccountInfo", url: str, job_data: JobFile, access_token: str
) -> Response:
    """Sends the job file to the remote server

    Args:
        account: the account details for accessing the API
        url: the URL to send the job file to
        job_data: the data of the job
        access_token: the temporary JWT token used for submitting jobs and downloading logfiles

    Returns:
        the response after the submission

    Raises:
        RuntimeError: Failed to POST job '{job_id}': {detail}
    """
    path = Path(tempfile.gettempdir()) / str(uuid4())

    try:
        # dump json to temporary file
        with path.open("w") as dest:
            job_data_json = job_data.model_dump_json(indent=2)
            dest.write(job_data_json)

        # Send temporary file to url
        with path.open("rb") as src:
            headers = {"Authorization": f"Bearer {access_token}"}
            response = requests.post(url, files={"upload_file": src}, headers=headers)

            # FIXME: Can the backend update the MSS when a job is queued i.e. after this request?
            if not response.ok:
                error_msg = _get_err_text(response)
                raise RuntimeError(
                    f"Failed to POST job '{job_data.job_id}': {error_msg}"
                )

    finally:
        # Delete temporary file
        path.unlink()

    return response


def cancel_job(upload_url: str, job_id: str, access_token: str) -> Response:
    """Cancels the job on the remote server

    Args:
        upload_url: the URL where the job file was sent to
        job_id: the unique identifier of the job
        access_token: the JWT token for accessing the backend

    Returns:
        the response after the submission

    Raises:
        RuntimeError: Failed to cancel job '{job_id}': {detail}
    """
    url = f"{upload_url}/{job_id}/cancel"
    headers = {"Authorization": f"Bearer {access_token}"}
    resp = requests.post(url, json={}, headers=headers)
    if not resp.ok:
        error_msg = _get_err_text(resp)
        raise RuntimeError(f"Failed to cancel job '{job_id}': {error_msg}")

    return resp


def download_job_logfile(job_id: str, url: str, access_token: str) -> Path:
    """Downloads the job logfile and returns the path to the downloaded file

    Args:
        job_id: the id of the job
        url: the URL to download from
        access_token: the JWT access token for accessing the backend

    Returns:
        the path to the downloaded job file

    Raises:
        RuntimeError: Failed to GET logfile of job '{job_id}': {detail}
    """
    headers = {"Authorization": f"Bearer {access_token}"}
    file_response = requests.get(url, stream=True, headers=headers)
    if not file_response.ok:
        error_msg = _get_err_text(file_response)
        raise RuntimeError(f"Failed to GET logfile of job '{job_id}': {error_msg}")

    job_logfile = Path(tempfile.gettempdir()) / (job_id + ".hdf5")
    with open(job_logfile, "wb") as file:
        shutil.copyfileobj(file_response.raw, file)

    return job_logfile


def get_remote_job_data(account: "AccountInfo", job_id: str) -> RemoteJob:
    """Retrieves the job data from the remote API

    Args:
        account: the account details for accessing the API
        job_id: the ID of the job to retrieve

    Returns:
        the dict representation of the job in the remote API

    Raises:
        RuntimeError: error retrieving job data: {detail}
        ValidationError: response is not a valid RemoteJob instance
    """
    url = f"{account.url}/jobs/{job_id}"
    headers = _get_auth_headers(account)
    response = requests.get(url, headers=headers)
    if response.ok:
        raw_data = response.json()
        return RemoteJob.model_validate(raw_data)

    error_msg = _get_err_text(response)
    raise RuntimeError(f"error retrieving job data: {error_msg}")


def _get_err_text(response: Response) -> str:
    """Returns the error text of the response

    Args:
        response: the response whose detail is to be returned

    Returns:
        the response error detail from the response
    """
    try:
        return response.json()["detail"]
    except (KeyError, AttributeError, JSONDecodeError):
        return response.text


def _get_auth_headers(provider_account: "AccountInfo") -> Optional[Dict[str, str]]:
    """Retrieves the auth header for this provider.

    Returns:
        dict of authorization of the authorization headers if account has auth else None
    """
    if provider_account.token is not None:
        return {"Authorization": f"Bearer {provider_account.token}"}

    return None
